-- hacer que solo se pueda ejecutar como extensión
\echo Utiliza "CREATE EXTENSION algoritmos_arbol CASCADE" para cargar este archivo. \quit

-- Crear schema para guardar las tablas de entrenamiento
CREATE SCHEMA algoritmos_arbol_datos;

CREATE TABLE nodos
(
  id                              INTEGER, -- Id del Nodo (-1 si es la raiz)
  id_padre                        INTEGER, -- Id del nodo padre
  tipo                            VARCHAR, -- tipo de nodo: RAIZ, RAMA, HOJA
  atributo                        VARCHAR, -- atributo de la rama (-1 si es la raiz)
  comparador                      VARCHAR, -- comparador (=)
  atributo_valor                  VARCHAR, -- valor del atributo de la rama (-1 si es la raiz)
  clase                           VARCHAR, -- clase elegida (-1 si no es una HOJA)
  instancias_cantidad_correctas   INTEGER, -- cantidad de instancias clasificadas correctamente
  instancias_cantidad_incorrectas INTEGER  -- cantidad de instancias clasificadas correctamente
);

CREATE TABLE IF NOT EXISTS info
(
  id     VARCHAR,
  titulo VARCHAR,
  valor  VARCHAR
);

CREATE TABLE IF NOT EXISTS textos_evaluacion
(
  id     VARCHAR,
  valor  TEXT
);


CREATE TABLE IF NOT EXISTS estadisticas_evaluacion
(
  tipo       VARCHAR, -- PARCIAL, TOTAL
  clase      VARCHAR,
  tpr        REAL,
  fpr        REAL,
  presicion  REAL,
  recall     REAL,
  f_measure  REAL,
  mcc        REAL,
  roc_area   REAL,
  prc_area   REAL
);

CREATE TABLE IF NOT EXISTS info_evaluacion
(
  id     VARCHAR,
  titulo VARCHAR,
  valor  REAL
);

CREATE TABLE matriz_confusion
(
  fila     VARCHAR,
  columna  VARCHAR,
  cantidad INTEGER
);

CREATE TABLE clases_probabilidades(
  nodo         INTEGER,
  clase        VARCHAR,
  probabilidad REAL
);

CREATE TABLE cantidades_anteriores(
  clase    VARCHAR,
  cantidad INTEGER
);

CREATE TABLE clasificados(
  id       SERIAL PRIMARY KEY,
  clase    VARCHAR,
  cantidad INTEGER
);

CREATE TABLE clasificados_probabilidades(
  id           INTEGER,
  clase        VARCHAR,
  probabilidad REAL,
  FOREIGN KEY (id) REFERENCES clasificados(id) ON DELETE CASCADE
);

CREATE OR REPLACE FUNCTION evaluar_tabla(tabla VARCHAR, atributo VARCHAR)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
BEGIN
  PERFORM algoritmos_arbol.evaluar_matriz(tabla, atributo, FALSE);
  PERFORM algoritmos_arbol.crear_estadisticas();
  RETURN 'OK'::VARCHAR;
END
$$;

CREATE OR REPLACE FUNCTION mostrar()
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
BEGIN
  PERFORM algoritmos_arbol.mostrar_arbol();
  PERFORM algoritmos_arbol.mostrar_info_arbol();
  PERFORM algoritmos_arbol.mostrar_info_evaluacion();
  PERFORM algoritmos_arbol.mostrar_evaluacion();
  RETURN 'OK'::VARCHAR;
END
$$;


CREATE OR REPLACE FUNCTION id3(tabla VARCHAR, atributo VARCHAR)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
BEGIN
    PERFORM algoritmos_arbol.id3(tabla, atributo, 'public');

    RETURN 'OK'::VARCHAR;
END
$$;

CREATE OR REPLACE FUNCTION id3(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
BEGIN
    PERFORM algoritmos_arbol.crear_arbol_id3(tabla, atributo, esquema);

    PERFORM algoritmos_arbol.evaluar_tabla(tabla, esquema);

    PERFORM algoritmos_arbol.mostrar();

    RETURN 'OK'::VARCHAR;
END
$$;


CREATE OR REPLACE FUNCTION id3_con_set(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, tabla_pruebas VARCHAR, esquema_pruebas VARCHAR)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
BEGIN
    PERFORM algoritmos_arbol.crear_arbol_id3(tabla, atributo, esquema);

    PERFORM algoritmos_arbol.evaluar_tabla(tabla_pruebas, esquema_pruebas);

    PERFORM algoritmos_arbol.mostrar();

    RETURN 'OK'::VARCHAR;
END
$$;


CREATE OR REPLACE FUNCTION id3_crossfold_validation(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, numero_folds INTEGER)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
DECLARE
    parametros TEXT;
BEGIN
    parametros := ''

    PERFORM algoritmos_arbol.cross_validation(tabla, atributo, esquema, 'id3', parametros, numero_folds);

    PERFORM algoritmos_arbol.mostrar();

    RETURN 'OK'::VARCHAR;
END
$$;


CREATE OR REPLACE FUNCTION id3_split(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR,  porcentaje_entrenamiento INTEGER)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
DECLARE
    entrenamiento TEXT;
    pruebas TEXT;
BEGIN
    entrenamiento := format('%s_%s_e%s', esquema, tabla, porcentaje_entrenamiento);
    pruebas := format('%s_%s_p%s', esquema, tabla, 100 - porcentaje_entrenamiento);

    PERFORM algoritmos_arbol.dividir_tabla(tabla, esquema, porcentaje_entrenamiento);

    PERFORM algoritmos_arbol.crear_arbol_id3(entrenamiento, atributo, 'algoritmos_arbol_datos');

    PERFORM algoritmos_arbol.evaluar_tabla(pruebas, 'algoritmos_arbol_datos');

    PERFORM algoritmos_arbol.mostrar();

    RETURN 'OK'::VARCHAR;
END
$$;

CREATE OR REPLACE FUNCTION c45(tabla VARCHAR, atributo VARCHAR)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
BEGIN
    PERFORM algoritmos_arbol.c45(tabla, atributo, 'public', 1, 0.25::float);

    RETURN 'OK'::VARCHAR;
END
$$;

CREATE OR REPLACE FUNCTION c45(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, objetos_minimos INTEGER, factor_confianza FLOAT)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
BEGIN
  PERFORM algoritmos_arbol.crear_arbol_c45(tabla, atributo, esquema, objetos_minimos, factor_confianza);

  PERFORM algoritmos_arbol.evaluar_tabla(tabla, esquema);

  PERFORM algoritmos_arbol.mostrar();

  RETURN 'OK'::VARCHAR;
END
$$;


CREATE OR REPLACE FUNCTION c45_con_set(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, objetos_minimos INTEGER, factor_confianza FLOAT, tabla_pruebas VARCHAR, esquema_pruebas VARCHAR)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
BEGIN
    PERFORM algoritmos_arbol.crear_arbol_c45(tabla, atributo, esquema, objetos_minimos, factor_confianza);

    PERFORM algoritmos_arbol.evaluar_tabla(tabla_pruebas, esquema_pruebas);

    PERFORM algoritmos_arbol.mostrar();

    RETURN 'OK'::VARCHAR;
END
$$;


CREATE OR REPLACE FUNCTION c45_crossfold_validation(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, objetos_minimos INTEGER, factor_confianza FLOAT, numero_folds INTEGER)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
DECLARE
    parametros TEXT;
BEGIN
    parametros := format(', %s, %s', objetos_minimos, factor_confianza);

    PERFORM algoritmos_arbol.cross_validation(tabla, atributo, esquema, 'c45', parametros, numero_folds);

    PERFORM algoritmos_arbol.mostrar();

    RETURN 'OK'::VARCHAR;
END
$$;


CREATE OR REPLACE FUNCTION c45_split(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, objetos_minimos INTEGER, factor_confianza FLOAT, porcentaje_entrenamiento INTEGER)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
DECLARE
    entrenamiento TEXT;
    pruebas TEXT;
BEGIN
    entrenamiento := format('%s_%s_e%s', esquema, tabla, porcentaje_entrenamiento);
    pruebas := format('%s_%s_p%s', esquema, tabla, 100 - porcentaje_entrenamiento);

    PERFORM algoritmos_arbol.dividir_tabla(tabla, esquema, porcentaje_entrenamiento);

    PERFORM algoritmos_arbol.crear_arbol_c45(entrenamiento, atributo, 'algoritmos_arbol_datos', objetos_minimos, factor_confianza);

    PERFORM algoritmos_arbol.evaluar_tabla(pruebas, 'algoritmos_arbol_datos');

    PERFORM algoritmos_arbol.mostrar();

    RETURN 'OK'::VARCHAR;
END
$$;


CREATE OR REPLACE FUNCTION
dividir_tabla_con_nombre(tabla VARCHAR, esquema VARCHAR, porcentaje_entrenamiento INTEGER, esquema_entrenamiento VARCHAR, tabla_entrenamiento VARCHAR, esquema_pruebas VARCHAR, tabla_pruebas VARCHAR, sobreescribir BOOLEAN)
RETURNS VARCHAR
STRICT
LANGUAGE plpython3u
AS
$$
# tabla = "tabla"
# esquema = "esquema"
# porcentaje_entrenamiento = 70
# esquema_entrenamiento = "esquema"
# tabla_entrenamiento = "tabla"
# esquema_pruebas = "esquema"
# tabla_pruebas = "tabla"
# sobreescribir = True

query = f"SELECT COUNT(*) FROM {esquema}.{tabla};"
total_filas = plpy.execute(query)[0]["count"]

cantidad_de_entrenamiento = int(total_filas * (porcentaje_entrenamiento / 100))
plpy.notice(f"cantidad de filas para de entrenamiento: {cantidad_de_entrenamiento}")

cantidad_pruebas = total_filas - cantidad_de_entrenamiento
plpy.notice(f"cantidad de filas para de pruebas: {cantidad_pruebas}")

if sobreescribir:
# Eliminar Tablas
    query = f"DROP TABLE IF EXISTS {esquema_pruebas}.{tabla_pruebas}"
    plpy.execute(query)

    query = f"DROP TABLE IF EXISTS {esquema_entrenamiento}.{tabla_entrenamiento}"
    plpy.execute(query)

# Crear la tabla de entrenamiento
query = f"""
CREATE TABLE {esquema_entrenamiento}.{tabla_entrenamiento} AS
SELECT * 
FROM {esquema}.{tabla}
"""
plpy.execute(query)


# Añadir id temporal
query = f"""
ALTER TABLE {esquema_entrenamiento}.{tabla_entrenamiento}
ADD COLUMN _id_temporal_pruebas INTEGER;
"""
plpy.execute(query)

# Darle valor al id temporal
query = f"""
WITH numbered_rows AS (
  SELECT *, ROW_NUMBER() OVER () AS row_num, ctid
  FROM {esquema_entrenamiento}.{tabla_entrenamiento}
)
UPDATE {esquema_entrenamiento}.{tabla_entrenamiento} AS e
SET _id_temporal_pruebas = nr.row_num
FROM numbered_rows nr
WHERE e.ctid = nr.ctid;
"""
plpy.execute(query)

# crear tabla de pruebas
query = f"""
CREATE TABLE {esquema_pruebas}.{tabla_pruebas} AS
SELECT * 
FROM {esquema_entrenamiento}.{tabla_entrenamiento}
ORDER BY RANDOM()
LIMIT {cantidad_pruebas};
"""
plpy.execute(query)

# Eliminar datos de la tabla de pruebas
# Que estén en la tabla de entrenamiento
query = f"""
DELETE FROM {esquema_entrenamiento}.{tabla_entrenamiento}
WHERE _id_temporal_pruebas IN (
    SELECT _id_temporal_pruebas
    FROM {esquema_pruebas}.{tabla_pruebas}
)
"""
plpy.execute(query)

# Eliminar columna de id temporal
query = f"""
ALTER TABLE {esquema_entrenamiento}.{tabla_entrenamiento}
DROP COLUMN _id_temporal_pruebas
"""
plpy.execute(query)

query = f"""
ALTER TABLE {esquema_pruebas}.{tabla_pruebas}
DROP COLUMN _id_temporal_pruebas
"""
plpy.execute(query)


# Eliminar Tablas

plpy.notice(f"Se creó la tabla {esquema_entrenamiento}.{tabla_entrenamiento}")
plpy.notice(f"Se creó la tabla {esquema_pruebas}.{tabla_pruebas}")

return 'OK'
$$;
            
CREATE OR REPLACE FUNCTION
dividir_tabla(tabla VARCHAR, esquema VARCHAR, porcentaje_entrenamiento INTEGER)
RETURNS VARCHAR
STRICT
LANGUAGE plpython3u
AS
$$
# tabla = "tabla"
# esquema = "esquema"
# porcentaje_entrenamiento = 70

# Nombres_de_tablas
entrenamiento = f"{esquema}_{tabla}_e{porcentaje_entrenamiento}"
pruebas = f"{esquema}_{tabla}_p{100 - porcentaje_entrenamiento}"

query = f"""
    SELECT * FROM algoritmos_arbol.dividir_tabla_con_nombre(
        '{tabla}',
        '{esquema}',
        '{porcentaje_entrenamiento}',
        'algoritmos_arbol_datos',
        '{entrenamiento}',
        'algoritmos_arbol_datos',
        '{pruebas}',
        TRUE
    )
    """

plpy.execute(query)

return 'OK'
$$;
            
CREATE OR REPLACE FUNCTION
crear_arbol_id3(tabla VARCHAR, atributo_objetivo VARCHAR, esquema VARCHAR)
RETURNS VARCHAR
STRICT
LANGUAGE plpython3u
AS
$$
# tabla = "test"
# atributo_objetivo = "test2"
# esquema = "public"

from collections import defaultdict
from typing import Dict, List, Literal, Optional, Tuple
from math import log2

# Eliminar arbol anterior
query = "TRUNCATE algoritmos_arbol.nodos;"
plpy.execute(query)
query = "TRUNCATE algoritmos_arbol.info;"
plpy.execute(query)

# Obtener clases posibles
query = f"""
    SELECT DISTINCT {atributo_objetivo} as clase
    FROM {esquema}.{tabla}
"""
resultado = plpy.execute(query)

clases_unicas: Dict[str, float] = {fila["clase"]: 0 for fila in resultado}


class Probabilidad:
    probabilidades: Dict[int, Dict[str, float]] = defaultdict(lambda: clases_unicas.copy())

    @staticmethod
    def actualizar_probabilidad(nodo: int, clase: str, probabilidad: float):
        Probabilidad.probabilidades[nodo][clase] = probabilidad

    @staticmethod
    def adicionar_en_tabla():
        query = "TRUNCATE TABLE algoritmos_arbol.clases_probabilidades"
        plpy.execute(query)

        values = []

        for nodo, clases in Probabilidad.probabilidades.items():
            for clase, probabilidad in clases.items():
                values.append(f"({nodo}, '{clase}', {probabilidad})")

        query = f"""
            INSERT INTO algoritmos_arbol.clases_probabilidades
            (nodo, clase, probabilidad) VALUES
            {",".join(values)}
        """
        plpy.execute(query)

    @staticmethod
    def eliminar_probabilidades_sueltas(id_nodos: List[int]):
        ids = list(Probabilidad.probabilidades.keys())
        for id in ids:
            if id not in id_nodos:
                del Probabilidad.probabilidades[id]

def es_numerico(columna: str) -> bool:
    query = f"""
    SELECT data_type AS tipo
    FROM information_schema.columns
    WHERE table_name = '{tabla}'
    AND table_schema = '{esquema}'
    AND column_name = '{columna}';
    """
    resultado = plpy.execute(query)
    tipo: str = resultado[0]["tipo"]
    return tipo in (
        "smallint",
        "integer",
        "bigint",
        "real",
        "double precision",
        "numeric",
    )


def entropia_clase(filtro: str = "true") -> Tuple[float, int]:
    query = f"SELECT count(*) as total FROM {esquema}.{tabla} WHERE {filtro}"

    resultado = plpy.execute(query)
    total_row = resultado[0]
    total_valor: int = total_row["total"]

    query = f"SELECT {atributo_objetivo}, count({atributo_objetivo}) as cuenta FROM {esquema}.{tabla} WHERE {filtro} GROUP BY {atributo_objetivo}"
    resultado = plpy.execute(query)

    entropia_valor = 0
    for fila in resultado:
        p_objetivo: float = fila["cuenta"] / total_valor
        entropia_valor -= p_objetivo * log2(p_objetivo)

    return entropia_valor, total_valor


def entropia_promedio(columna: str, filtro: str = "true") -> float:
    query = f"SELECT count(*) as total FROM {esquema}.{tabla} WHERE {filtro}"
    resultado = plpy.execute(query)
    total_row = resultado[0]
    total: int = total_row["total"]

    query = f"SELECT DISTINCT {columna} as valor FROM {esquema}.{tabla} WHERE {filtro}"
    resultado = plpy.execute(query)
    valores: List[str] = []
    for fila in resultado:
        valores.append(fila["valor"])

    entropia_total = 0
    for valor in valores:
        filtro_valor = f"({filtro}) AND {columna} = '{valor}'"
        entropia_valor, total_valor = entropia_clase(filtro_valor)

        entropia_total += (total_valor / total) * entropia_valor
    return entropia_total


class Nodo:
    id: int
    hijos: List["Nodo"]
    padre: Optional["Nodo"]
    tipo: Literal["RAIZ", "RAMA", "HOJA"]

    atributo: str
    atributo_valor: str

    clase: Optional[str]
    instancias_cantidad_correctas: int
    instancias_cantidad_incorrectas: int
    clase_total: int

    entropia: float

    atributos_disponibles: List[str]

    contador_id: int = 0

    instancias_clasificadas_incorrectamente: int = 0
    instancias_clasificadas_correctamente: int = 0

    hojas: int = 0

    atributos_nominales_con_valores: Dict[str, List[str]] = {}

    def __init__(
        self,
        padre: Optional["Nodo"],
        tipo: Literal["RAIZ", "RAMA", "HOJA"],
        atributo: str,
        atributo_valor: str,
        atributos_disponibles: List[str],
    ) -> None:
        self.padre = padre
        self.tipo = tipo
        self.atributo = atributo
        self.atributo_valor = atributo_valor
        self.atributos_disponibles = atributos_disponibles
        self.hijos = []

        self.clase = None
        self.instancias_cantidad_correctas = 0
        self.instancias_cantidad_incorrectas = 0
        self.clase_total = 0

        # calculo entropia
        entropia, _ = entropia_clase(self.crear_filtro())
        self.entropia = entropia

        self.id = Nodo.contador_id
        Nodo.contador_id += 1

        if tipo == "HOJA":
            Nodo.hojas += 1

    def crear_filtro(self) -> str:
        filtro = "true"
        nodo = self
        while nodo != None:
            if nodo.tipo == "RAMA" or nodo.tipo == "HOJA":
                filtro += f" AND {nodo.atributo} = '{nodo.atributo_valor}'"

            nodo = nodo.padre

        return filtro

    def dar_clase(self):
        query = f"""
            SELECT {atributo_objetivo} AS clase, count(*) AS cantidad 
            FROM {esquema}.{tabla}
            WHERE {self.crear_filtro()}
            GROUP BY {atributo_objetivo} ORDER BY cantidad DESC;
        """
        resultado = plpy.execute(query)

        self.clase = resultado[0]["clase"]

        cantidades = [fila["cantidad"] for fila in resultado]

        total = sum(cantidades)

        self.clase_total = total

        self.instancias_cantidad_correctas = cantidades[0]
        Nodo.instancias_clasificadas_correctamente += self.instancias_cantidad_correctas

        self.instancias_cantidad_incorrectas = sum(cantidades[1:])
        Nodo.instancias_clasificadas_incorrectamente += (
            self.instancias_cantidad_incorrectas
        )

        for fila in resultado:
            Probabilidad.actualizar_probabilidad(
                self.id, fila["clase"], fila["cantidad"] / total
            )

    def crear_hijos(self):
        # elegir mejor atributo
        filtro = self.crear_filtro()
        ganancias: Dict[str, float] = {}
        for attr in self.atributos_disponibles:
            if es_numerico(attr):
                plpy.error(
                    f"La columna {attr} es de tipo númerico, "
                    + " no es compatible con este algoritmo"
                )
            ganancias[attr] = self.entropia - entropia_promedio(attr, filtro)

        mejor_ganancia: str = max(ganancias, key=lambda k: ganancias[k])

        query = f"""
            SELECT atributo_valor, count(*) AS contador
            FROM (SELECT {mejor_ganancia} AS atributo_valor, {atributo_objetivo}
                  FROM {esquema}.{tabla}
                  WHERE {filtro}
                  GROUP BY {mejor_ganancia}, {atributo_objetivo}) AS grupo
            GROUP BY atributo_valor ORDER BY contador;
        """
        resultado = plpy.execute(query)

        atributos_disponibles_hijo = self.atributos_disponibles[::]
        atributos_disponibles_hijo.remove(mejor_ganancia)

        # {atributo_valor: str, contador: int}
        for fila in resultado:
            if fila["contador"] <= 1 or len(atributos_disponibles_hijo) == 0:
                hijo = Nodo(
                    self,
                    "HOJA",
                    mejor_ganancia,
                    fila["atributo_valor"],
                    atributos_disponibles_hijo,
                )
                hijo.dar_clase()
            else:
                hijo = Nodo(
                    self,
                    "RAMA",
                    mejor_ganancia,
                    fila["atributo_valor"],
                    atributos_disponibles_hijo,
                )
                hijo.crear_hijos()

            self.hijos.append(hijo)

    def obtener_profundidad(self) -> int:
        profundidad = 0
        nodo = self.padre
        while nodo != None:
            profundidad += 1
            nodo = nodo.padre

        return profundidad

    def adicionar_nodos_en_tablas(self):
        id = self.id
        padre = self.padre.id if self.padre != None else -1
        tipo = self.tipo
        atributo = self.atributo
        atributo_valor = self.atributo_valor
        clase = "NULL" if self.clase is None else f"'{self.clase}'"
        instancias_cantidad_correctas = self.instancias_cantidad_correctas
        instancias_cantidad_incorrectas = self.instancias_cantidad_incorrectas

        plpy.execute(
            """
            INSERT INTO algoritmos_arbol.nodos
            (
                id,
                id_padre,
                tipo,
                atributo,
                comparador,
                atributo_valor,
                clase,
                instancias_cantidad_correctas,
                instancias_cantidad_incorrectas
            )
            VALUES (%s, %s, '%s', '%s', '%s', '%s', %s, %s, %s)
        """
            % (
                id,
                padre,
                tipo,
                atributo,
                "=",
                atributo_valor,
                clase,
                instancias_cantidad_correctas,
                instancias_cantidad_incorrectas,
            )
        )

        for hijo in self.hijos:
            hijo.adicionar_nodos_en_tablas()

    def obtener_id_nodos(self, lista: List[int]):
        lista.append(self.id)
        for hijo in self.hijos:
            hijo.obtener_id_nodos(lista)

    def cantidad_nodos(self):
        cantidad = 1
        for hijo in self.hijos:
            cantidad += hijo.cantidad_nodos()

        return cantidad

    def cantidad_hojas(self):
        if self.tipo == "HOJA":
            cantidad = 1
        else:
            cantidad = 0
        for hijo in self.hijos:
            cantidad += hijo.cantidad_hojas()

        return cantidad

    def obtener_cantidad_clase(self, cuenta: Dict[str, int]):
        if self.tipo == "HOJA" and self.clase is not None:
            cuenta[self.clase] += self.instancias_cantidad_correctas

        for hijo in self.hijos:
            hijo.obtener_cantidad_clase(cuenta)

    def obtener_clase_mas_probable(self):
        cantidad_clases: Dict[str, int] = defaultdict(int)
        self.obtener_cantidad_clase(cantidad_clases)
        return max(cantidad_clases, key=lambda k: cantidad_clases[k])


    def completar_atributos(self):
        if len(self.hijos) == 0:
            return

        for hijo in self.hijos:
            hijo.completar_atributos()

        atributo = self.hijos[0].atributo

        valores_hijos = [hijo.atributo_valor for hijo in self.hijos]
        valores_faltanes: List[str] = []

        if atributo not in Nodo.atributos_nominales_con_valores:
            return

        for valor in Nodo.atributos_nominales_con_valores[atributo]:
            if not valor in valores_hijos:
                valores_faltanes.append(valor)

        if len(valores_faltanes) == 0:
            return

        clase_mas_probable = self.obtener_clase_mas_probable()

        for valor in valores_faltanes:
            nodo = Nodo(self, "HOJA", atributo, valor, [])
            nodo.clase = clase_mas_probable
            Probabilidad.actualizar_probabilidad(nodo.id, nodo.clase, 0)
            self.hijos.append(nodo)

    def __repr__(self) -> str:
        return f"<Nodo tipo={self.tipo} atributo={self.atributo} atributo_valor={self.atributo_valor} entropia={self.entropia}>"


query = f"""
SELECT column_name as columna 
FROM information_schema.columns 
WHERE table_name = '{tabla}' 
AND table_schema = '{esquema}';
"""

resultado = plpy.execute(query)

atributos: List[str] = []

for fila in resultado:
    if fila["columna"] != atributo_objetivo:
        atributos.append(fila["columna"])

raiz = Nodo(None, "RAIZ", "-1", "-1", atributos)
raiz.crear_hijos()

# obtener atributos usados en cada nodo
atributos_nominales_usados: set[str] = set()


def agregar_atributos_usados(nodo: Nodo):
    if nodo.tipo != "RAIZ":
        atributos_nominales_usados.add(nodo.atributo)

    for hijo in nodo.hijos:
        agregar_atributos_usados(hijo)


agregar_atributos_usados(raiz)

for atributo in atributos_nominales_usados:
    query = f"""
        SELECT DISTINCT {atributo} as atributo
        FROM {esquema}.{tabla}
    """
    resultado = plpy.execute(query)
    valores = [fila["atributo"] for fila in resultado]
    Nodo.atributos_nominales_con_valores[atributo] = valores

raiz.completar_atributos()

raiz.adicionar_nodos_en_tablas()

ids = []
raiz.obtener_id_nodos(ids)
#Probabilidad.eliminar_probabilidades_sueltas(ids)
Probabilidad.adicionar_en_tabla()

def insertar_info(id, titulo, valor):
    plpy.execute(
        """
        INSERT INTO algoritmos_arbol.info (id, titulo, valor)
        VALUES ('%s', '%s', '%s')
    """
        % (id, titulo, valor)
    )


insertar_info("tipo", "Algoritmo Usado", "ID3")
insertar_info("tabla", "Tabla", tabla)
insertar_info("attr_objetivo", "Atributo Objetivo", atributo_objetivo)
insertar_info("esquema", "Esquema", esquema)
insertar_info("num_nodos", "Número de Nodos", raiz.cantidad_nodos())
insertar_info("num_hojas", "Número de Hojas", raiz.cantidad_hojas())

return 'OK'
$$;
            
CREATE OR REPLACE FUNCTION
crear_arbol_c45(tabla VARCHAR, atributo_objetivo VARCHAR, esquema VARCHAR, objetos_minimos INTEGER, factor_confianza FLOAT)
RETURNS VARCHAR
STRICT
LANGUAGE plpython3u
AS
$$
# tabla = "test"
# atributo_objetivo = "test2"
# esquema = "public"
# objetos_minimos = 1
# factor_confianza = 0.25
colapsar_arbol = True
elevacion_subarbol = True

import math
from typing import Dict, List, Literal, Optional, Tuple
import functools
from collections import defaultdict

# Eliminar arbol anterior
query = "TRUNCATE algoritmos_arbol.nodos;"
plpy.execute(query)
query = "TRUNCATE algoritmos_arbol.info;"
plpy.execute(query)


class Utils:
    SMALL: float = 1e-6

    @staticmethod
    def smOrEq(a: float, b: float) -> bool:
        return (a - b < Utils.SMALL) or (a <= b)

    @staticmethod
    def grOrEq(a: float, b: float) -> bool:
        return (b - a < Utils.SMALL) or (a >= b)

    @staticmethod
    def sm(a: float, b: float) -> bool:
        return b - a > Utils.SMALL

    @staticmethod
    def gr(a: float, b: float) -> bool:
        return a - b > Utils.SMALL

    @staticmethod
    def eq(a: float, b: float) -> bool:
        return (a == b) or ((a - b < Utils.SMALL) and (b - a < Utils.SMALL))


class Estadistica:
    P0 = (
        -5.99633501014107895267e1,
        9.80010754185999661536e1,
        -5.66762857469070293439e1,
        1.39312609387279679503e1,
        -1.23916583867381258016e0,
    )
    Q0 = (
        1.95448858338141759834e0,
        4.67627912898881538453e0,
        8.63602421390890590575e1,
        -2.25462687854119370527e2,
        2.00260212380060660359e2,
        -8.20372256168333339912e1,
        1.59056225126211695515e1,
        -1.18331621121330003142e0,
    )

    P1 = (
        4.05544892305962419923e0,
        3.15251094599893866154e1,
        5.71628192246421288162e1,
        4.40805073893200834700e1,
        1.46849561928858024014e1,
        2.18663306850790267539e0,
        -1.40256079171354495875e-1,
        -3.50424626827848203418e-2,
        -8.57456785154685413611e-4,
    )
    Q1 = (
        1.57799883256466749731e1,
        4.53907635128879210584e1,
        4.13172038254672030440e1,
        1.50425385692907503408e1,
        2.50464946208309415979e0,
        -1.42182922854787788574e-1,
        -3.80806407691578277194e-2,
        -9.33259480895457427372e-4,
    )

    P2 = (
        3.23774891776946035970e0,
        6.91522889068984211695e0,
        3.93881025292474443415e0,
        1.33303460815807542389e0,
        2.01485389549179081538e-1,
        1.23716634817820021358e-2,
        3.01581553508235416007e-4,
        2.65806974686737550832e-6,
        6.23974539184983293730e-9,
    )
    Q2 = (
        6.02427039364742014255e0,
        3.67983563856160859403e0,
        1.37702099489081330271e0,
        2.16236993594496635890e-1,
        1.34204006088543189037e-2,
        3.28014464682127739104e-4,
        2.89247864745380683936e-6,
        6.79019408009981274425e-9,
    )

    @staticmethod
    def addErrores(
        instancias: float, error_observado: float, valor_confianza: float
    ) -> float:
        n = instancias
        e = error_observado
        cf = valor_confianza

        if cf > 0.5:
            return 0

        if e < 1:
            base = n * (1 - math.pow(cf, 1 / n))

            if e == 0:
                return base
            else:
                return base + e * (Estadistica.addErrores(n, 1, cf) - base)

        if e + 0.5 >= n:
            return max(n - e, 0)

        z = Estadistica.normalInversa(1 - cf)

        f = (e + 0.5) / n
        r = (
            f
            + (z * z) / (2 * n)
            + z * math.sqrt((f / n) - (f * f / n) + (z * z / (4 * n * n)))
        ) / (1 + (z * z) / n)

        return (r * n) - e

    @staticmethod
    def p1evl(x, coef, N):
        ans = x + coef[0]
        for i in range(1, N):
            ans = ans * x + coef[i]
        return ans

    @staticmethod
    def polevl(x, coef, N):
        ans = coef[0]
        for i in range(1, N + 1):
            ans = ans * x + coef[i]
        return ans

    @staticmethod
    def normalInversa(y0: float):
        s2pi = math.sqrt(2.0 * math.pi)

        if y0 <= 0.0:
            raise ValueError("El argumento debe ser positivo")
        if y0 >= 1.0:
            raise ValueError("El argumento debe ser menor que 1")

        code = 1
        y = y0

        if y > (1.0 - 0.13533528323661269189):  # 0.135... = exp(-2)
            y = 1.0 - y
            code = 0

        if y > 0.13533528323661269189:
            y = y - 0.5
            y2 = y * y
            x = y + y * (
                y2
                * Estadistica.polevl(y2, Estadistica.P0, 4)
                / Estadistica.p1evl(y2, Estadistica.Q0, 8)
            )
            x = x * s2pi
            return x

        x = math.sqrt(-2.0 * math.log(y))
        x0 = x - math.log(x) / x
        z = 1.0 / x

        if x < 8.0:
            x1 = (
                z
                * Estadistica.polevl(z, Estadistica.P1, 8)
                / Estadistica.p1evl(z, Estadistica.Q1, 8)
            )
        else:
            x1 = (
                z
                * Estadistica.polevl(z, Estadistica.P2, 8)
                / Estadistica.p1evl(z, Estadistica.Q2, 8)
            )

        x = x0 - x1

        if code != 0:
            x = -x

        return x


# Obtener clases posibles
query = f"""
    SELECT DISTINCT {atributo_objetivo} as clase
    FROM {esquema}.{tabla}
"""
resultado = plpy.execute(query)

clases_unicas: Dict[str, float] = {fila["clase"]: 0 for fila in resultado}


class Probabilidad:
    probabilidades: Dict[int, Dict[str, float]] = defaultdict(
        lambda: clases_unicas.copy()
    )

    @staticmethod
    def actualizar_probabilidad(nodo: int, clase: str, probabilidad: float):
        Probabilidad.probabilidades[nodo][clase] = probabilidad

    @staticmethod
    def adicionar_en_tabla():
        query = "TRUNCATE TABLE algoritmos_arbol.clases_probabilidades"
        plpy.execute(query)

        values = []

        for nodo, clases in Probabilidad.probabilidades.items():
            for clase, probabilidad in clases.items():
                values.append(f"({nodo}, '{clase}', {probabilidad})")

        query = f"""
            INSERT INTO algoritmos_arbol.clases_probabilidades
            (nodo, clase, probabilidad) VALUES
            {",".join(values)}
        """
        plpy.execute(query)

    @staticmethod
    def eliminar_probabilidades_sueltas(id_nodos: List[int]):
        ids = list(Probabilidad.probabilidades.keys())
        for id in ids:
            if id not in id_nodos:
                del Probabilidad.probabilidades[id]


@functools.lru_cache(maxsize=None)
def entropia_clase(filtro: str = "true") -> Tuple[float, int]:
    query = f"""
        SELECT {atributo_objetivo}, count({atributo_objetivo}) as cuenta
        FROM {esquema}.{tabla} 
        WHERE {filtro} 
        GROUP BY {atributo_objetivo}
    """

    resultado = plpy.execute(query)
    if not resultado:
        return 0.0, 0

    total_valor = sum(fila["cuenta"] for fila in resultado)
    entropia_valor = sum(
        -fila["cuenta"] / total_valor * math.log2(fila["cuenta"] / total_valor)
        for fila in resultado
    )
    return entropia_valor, total_valor


def entropia_promedio(columna: str, filtro: str = "true") -> float:
    query = f"""
    SELECT {columna} as valor, {atributo_objetivo}, COUNT(*) as cuenta 
    FROM {esquema}.{tabla} 
    WHERE {filtro}
    GROUP BY {columna}, {atributo_objetivo}
    """
    resultado = plpy.execute(query)

    total = sum(row["cuenta"] for row in resultado)

    counts = {}
    for row in resultado:
        valor = row["valor"]
        objetivo = row[atributo_objetivo]
        if valor not in counts:
            counts[valor] = {}
        counts[valor][objetivo] = row["cuenta"]

    entropia_total = 0
    for valor, objetivos in counts.items():
        total_valor = sum(objetivos.values())
        entropia_valor = 0
        for count in objetivos.values():
            p_objetivo = count / total_valor
            entropia_valor -= p_objetivo * math.log2(p_objetivo)
        entropia_total -= (total_valor / total) * entropia_valor

    return -entropia_total


def calcular_mediana(columna: str, filtro: str = "true") -> float | int:
    query = f"SELECT count({columna}) AS cantidad FROM {esquema}.{tabla} WHERE {filtro}"
    resultado = plpy.execute(query)
    cantidad = resultado[0]["cantidad"]

    if cantidad % 2 == 0:
        query = f"""
            SELECT {columna} AS columna
            FROM {esquema}.{tabla}
            ORDER BY {columna}
            limit 2 OFFSET ({cantidad/2} - 1);
        """
        resultado = plpy.execute(query)
        suma = resultado[0]["columna"] + resultado[1]["columna"]
        mediana = suma / 2
    else:
        query = f"""
            SELECT {columna} AS columna
            FROM {esquema}.{tabla}
            ORDER BY {columna}
            limit 1 OFFSET ({math.floor(cantidad/2)});
        """
        resultado = plpy.execute(query)
        mediana = resultado[0]["columna"]

    return mediana


def entropia_promedio_numerico(columna: str, filtro: str = "true") -> float:
    query = f"SELECT count(*) as total FROM {esquema}.{tabla} WHERE {filtro}"
    resultado = plpy.execute(query)
    total_row = resultado[0]
    total: int = total_row["total"]

    mediana = calcular_mediana(columna, filtro)

    entropia_total: float = 0

    # <=
    filtro_valor = f"({filtro}) AND {columna} <= '{mediana}'"
    entropia_valor, total_valor = entropia_clase(filtro_valor)

    entropia_total += (total_valor / total) * entropia_valor

    # >
    filtro_valor = f"({filtro}) AND {columna} > '{mediana}'"
    entropia_valor, total_valor = entropia_clase(filtro_valor)

    entropia_total += (total_valor / total) * entropia_valor

    return entropia_total


@functools.lru_cache(maxsize=None)
def es_numerico(columna: str) -> bool:
    query = f"""
    SELECT data_type AS tipo
    FROM information_schema.columns
    WHERE table_name = '{tabla}' 
    AND table_schema = '{esquema}'
    AND column_name = '{columna}';
    """
    resultado = plpy.execute(query)
    tipo: str = resultado[0]["tipo"]
    return tipo in (
        "smallint",
        "integer",
        "bigint",
        "real",
        "double precision",
        "numeric",
    )


@functools.lru_cache(maxsize=None)
def split_informacion(attr: str, filtro: str = "true") -> float:
    query = f"""
        SELECT {attr}, count(*) AS cantidad
        FROM {esquema}.{tabla} 
        WHERE {filtro}
        GROUP BY {attr}
    """
    resultado = plpy.execute(query)

    if not resultado:
        return 0.0

    total = sum(fila["cantidad"] for fila in resultado)
    split_total = sum(
        (cantidad / total) * math.log2(cantidad / total)
        for fila in resultado
        for cantidad in [fila["cantidad"]]
    )

    return -split_total


class InfoGanancia:
    atributo: str
    ganancia: float
    ratio_ganancia: float

    def __init__(self, atributo: str) -> None:
        self.atributo = atributo
        self.ganancia = 0
        self.ratio_ganancia = 0

    def __repr__(self) -> str:
        return f"<attr={self.atributo} ganancia={self.ganancia} ratio_ganancia={self.ratio_ganancia}>"


class Nodo:
    __slots__ = [
        "id",
        "hijos",
        "padre",
        "tipo",
        "atributo",
        "comparador",
        "atributo_valor",
        "clase",
        "instancias_cantidad_correctas",
        "instancias_cantidad_incorrectas",
        "cantidad_objetos",
        "entropia",
        "atributos_disponibles",
    ]

    id: int
    hijos: List["Nodo"]
    padre: Optional["Nodo"]
    tipo: Literal["RAIZ", "RAMA", "HOJA"]

    atributo: str
    comparador: Literal["=", "<=", ">"]
    atributo_valor: str

    clase: Optional[str]
    instancias_cantidad_correctas: int
    instancias_cantidad_incorrectas: int

    cantidad_objetos: int

    entropia: float

    atributos_disponibles: List[str]

    contador_id: int = 0

    atributos_nominales_con_valores: Dict[str, List[str]] = {}

    def __init__(
        self,
        padre: Optional["Nodo"],
        tipo: Literal["RAIZ", "RAMA", "HOJA"],
        atributo: str,
        comparador: Literal["=", "<=", ">"],
        atributo_valor: str,
        atributos_disponibles: List[str],
    ) -> None:
        self.padre = padre
        self.tipo = tipo
        self.atributo = atributo
        self.comparador = comparador
        self.atributo_valor = atributo_valor
        self.atributos_disponibles = atributos_disponibles
        self.hijos = []

        self.clase = None
        self.instancias_cantidad_correctas = 0
        self.instancias_cantidad_incorrectas = 0

        # calculo entropia
        entropia, cantidad_objetos = entropia_clase(self.crear_filtro())
        self.entropia = entropia
        self.cantidad_objetos = cantidad_objetos

        self.id = Nodo.contador_id
        Nodo.contador_id += 1

        cantidad = 10 ** (len(str(self.id)) - 1)
        if self.id % cantidad == 0:
            plpy.notice(f"se creó el nodo No {self.id}")

    def crear_filtro(self) -> str:
        filtro = "true"
        nodo = self
        while nodo != None:
            if (nodo.tipo == "RAMA" or nodo.tipo == "HOJA") and self.padre != None:
                filtro += (
                    f" AND {nodo.atributo} {nodo.comparador} '{nodo.atributo_valor}'"
                )

            nodo = nodo.padre

        return filtro

    def dar_clase(self):
        query = f"""
            SELECT {atributo_objetivo} AS clase, count(*) AS cantidad 
            FROM {esquema}.{tabla}
            WHERE {self.crear_filtro()}
            GROUP BY {atributo_objetivo} ORDER BY cantidad DESC;
        """
        resultado = plpy.execute(query)

        if self.tipo == "RAMA":
            self.tipo = "HOJA"
        self.hijos = []

        if len(resultado) == 0:
            self.instancias_cantidad_correctas = 0
            self.instancias_cantidad_incorrectas = 0
            # if self.padre is not None:
            #     self.clase = self.padre.obtener_clase_mas_probable()
            return

        self.clase = resultado[0]["clase"]

        cantidades = [fila["cantidad"] for fila in resultado]

        self.instancias_cantidad_correctas = cantidades[0]
        self.instancias_cantidad_incorrectas = sum(cantidades[1:])

        total = sum(cantidades)

        for fila in resultado:
            Probabilidad.actualizar_probabilidad(
                self.id, fila["clase"], fila["cantidad"] / total
            )

    def obtener_atributos_inutiles(self, filtro: str):
        atributos_ignorados: List[str] = []

        for atributo in self.atributos_disponibles:
            query = f"""
                SELECT count(*) AS cantidad
                FROM {esquema}.{tabla} WHERE {filtro}
                GROUP BY {atributo}
            """
            resultado = plpy.execute(query)

            validos = 0
            for fila in resultado:
                if fila["cantidad"] >= objetos_minimos:
                    validos += 1
                if validos > 1:
                    break

            if validos <= 1:
                atributos_ignorados.append(atributo)

        return atributos_ignorados

    def crear_hijos(self):
        # Prevenir que se parta mas el arbol
        if self.cantidad_objetos < 2 * objetos_minimos:
            self.dar_clase()
            return

        # elegir mejor atributo
        filtro = self.crear_filtro()

        atributos_ignorados = self.obtener_atributos_inutiles(filtro)

        if len(atributos_ignorados) == len(self.atributos_disponibles):
            self.dar_clase()
            return

        ganancia_promedio = 0

        # atributo: {"ganancia":... , "ratio_ganancia":...}
        informacion_ganancias: List[InfoGanancia] = []

        for atributo in self.atributos_disponibles:
            if atributo in atributos_ignorados:
                continue

            if es_numerico(atributo):
                ganancia = self.entropia - entropia_promedio_numerico(atributo, filtro)
                split = split_informacion(atributo, filtro)
            else:
                ganancia = self.entropia - entropia_promedio(atributo, filtro)
                split = split_informacion(atributo, filtro)

            info = InfoGanancia(atributo)
            info.ganancia = ganancia
            if ganancia > 0 and split != 0:
                info.ratio_ganancia = ganancia / split

            ganancia_promedio += ganancia
            informacion_ganancias.append(info)

        total_atributos = len(self.atributos_disponibles) - len(atributos_ignorados)
        ganancia_promedio = ganancia_promedio / total_atributos

        mejor_atributo = ""
        mejor_gain_ratio = 0

        for info in informacion_ganancias:
            ganancia = info.ganancia
            ratio_ganancia = info.ratio_ganancia
            atributo = info.atributo

            if (
                ganancia >= ganancia_promedio - 1e-3
                and ratio_ganancia > mejor_gain_ratio
            ):
                mejor_atributo = atributo
                mejor_gain_ratio = ratio_ganancia

        if mejor_gain_ratio == 0:
            self.dar_clase()
            return

        atributos_disponibles_hijo = self.atributos_disponibles[::]
        atributos_disponibles_hijo.remove(mejor_atributo)

        # plpy.notice(
        #    f"Gain Ratios ({self.id}): {sorted(informacion_ganancias, key=lambda k: k.ratio_ganancia, reverse=True)}"
        # )
        # plpy.notice(f"mejor: {mejor_gain_ratio}")
        # plpy.notice("-----------")

        if es_numerico(mejor_atributo):
            mediana = calcular_mediana(mejor_atributo, filtro)

            # Crear hijo para <=
            filtro_menor_igual = f"({filtro}) AND {mejor_atributo} <= {mediana}"
            query = f"""
            SELECT count(DISTINCT {atributo_objetivo}) AS contador
            FROM {esquema}.{tabla}
            WHERE {filtro_menor_igual}
            """
            resultado = plpy.execute(query)
            if resultado[0]["contador"] <= 1 or len(atributos_disponibles_hijo) == 0:
                hijo = Nodo(
                    self,
                    "HOJA",
                    mejor_atributo,
                    "<=",
                    str(mediana),
                    atributos_disponibles_hijo,
                )
                hijo.dar_clase()
            else:
                hijo = Nodo(
                    self,
                    "RAMA",
                    mejor_atributo,
                    "<=",
                    str(mediana),
                    atributos_disponibles_hijo,
                )

            self.hijos.append(hijo)

            # Crear hijo para >
            filtro_mayor = f"({filtro}) AND {mejor_atributo} > {mediana}"
            query = f"""
            SELECT count(DISTINCT {atributo_objetivo}) AS contador
            FROM {esquema}.{tabla}
            WHERE {filtro_mayor}
            """
            resultado = plpy.execute(query)
            if resultado[0]["contador"] <= 1 or len(atributos_disponibles_hijo) == 0:
                hijo = Nodo(
                    self,
                    "HOJA",
                    mejor_atributo,
                    ">",
                    str(mediana),
                    atributos_disponibles_hijo,
                )
                hijo.dar_clase()

            else:
                hijo = Nodo(
                    self,
                    "RAMA",
                    mejor_atributo,
                    ">",
                    str(mediana),
                    atributos_disponibles_hijo,
                )

            self.hijos.append(hijo)
        else:
            query = f"""
                SELECT atributo_valor, count(*) AS contador
                FROM (SELECT {mejor_atributo} AS atributo_valor, {atributo_objetivo}
                      FROM {esquema}.{tabla}
                      WHERE {filtro}
                      GROUP BY {mejor_atributo}, {atributo_objetivo}) AS grupo
                GROUP BY atributo_valor ORDER BY contador;
            """
            resultado = plpy.execute(query)

            # {atributo_valor: str, contador: int}
            for fila in resultado:
                if fila["contador"] <= 1 or len(atributos_disponibles_hijo) == 0:
                    hijo = Nodo(
                        self,
                        "HOJA",
                        mejor_atributo,
                        "=",
                        fila["atributo_valor"],
                        atributos_disponibles_hijo,
                    )
                    hijo.dar_clase()
                else:
                    hijo = Nodo(
                        self,
                        "RAMA",
                        mejor_atributo,
                        "=",
                        fila["atributo_valor"],
                        atributos_disponibles_hijo,
                    )

                self.hijos.append(hijo)

            cantidad_hijos_validos = 0
            for hijo in self.hijos:
                if hijo.cantidad_objetos >= objetos_minimos:
                    cantidad_hijos_validos += 1

                if cantidad_hijos_validos > 1:
                    break

            if cantidad_hijos_validos <= 1:
                plpy.notice(f"El nodo no tiene suficientes hijos validos: {self}")
                plpy.notice(
                    f"{ hijo.atributo }: {[(hijo.atributo_valor, hijo.cantidad_objetos) for hijo in self.hijos]}"
                )
                plpy.notice(self.crear_filtro())
                plpy.notice("-------------------------------------")
                self.dar_clase()
                return

            for hijo in self.hijos:
                if hijo.tipo == "RAMA":
                    hijo.crear_hijos()

    def adicionar_en_tabla(self):
        id = self.id
        padre = self.padre.id if self.padre != None else -1
        tipo = self.tipo
        atributo = self.atributo
        comparador = self.comparador
        atributo_valor = self.atributo_valor
        clase = "NULL" if self.clase is None else f"'{self.clase}'"
        instancias_cantidad_correctas = self.instancias_cantidad_correctas
        instancias_cantidad_incorrectas = self.instancias_cantidad_incorrectas

        plpy.execute(
            """
            INSERT INTO algoritmos_arbol.nodos
            (
                id,
                id_padre,
                tipo,
                atributo,
                comparador,
                atributo_valor,
                clase,
                instancias_cantidad_correctas,
                instancias_cantidad_incorrectas
            )
            VALUES (%s, %s, '%s', '%s', '%s', '%s', %s, %s, %s)
        """
            % (
                id,
                padre,
                tipo,
                atributo,
                comparador,
                atributo_valor,
                clase,
                instancias_cantidad_correctas,
                instancias_cantidad_incorrectas,
            )
        )

    def adicionar_hijos_en_tabla(self):
        for hijo in self.hijos:
            hijo.adicionar_en_tabla()

        for hijo in self.hijos:
            hijo.adicionar_hijos_en_tabla()

    def reevaluar(self) -> None:
        query = f"""
            SELECT count(*) AS cantidad 
            FROM {esquema}.{tabla}
            WHERE {self.crear_filtro()};
        """
        resultado = plpy.execute(query)
        self.cantidad_objetos = resultado[0]["cantidad"]

        if self.tipo == "HOJA":
            self.dar_clase()
        else:
            for hijo in self.hijos:
                hijo.reevaluar()

    def error_hijos(self) -> int:
        if self.tipo == "HOJA":
            return self.instancias_cantidad_incorrectas
        else:
            total = 0
            for hijo in self.hijos:
                total += hijo.error_hijos()
            return total

    def error_actual(self) -> int:
        query = f"""
            SELECT {atributo_objetivo} AS clase, count(*) AS cantidad 
            FROM {esquema}.{tabla}
            WHERE {self.crear_filtro()}
            GROUP BY {atributo_objetivo} ORDER BY cantidad DESC OFFSET 1;
        """
        resultado = plpy.execute(query)

        cantidades = [fila["cantidad"] for fila in resultado]

        return sum(cantidades)

    def colapsar(self):
        if self.tipo != "HOJA":
            error_hijos = self.error_hijos()
            error_actual = self.error_actual()

            if error_hijos >= error_actual - 1e-3:
                self.dar_clase()
            else:
                for hijo in self.hijos:
                    hijo.colapsar()

    def error_estimado(self, filtro: str):
        global factor_confianza

        query = f"""
            SELECT {atributo_objetivo} AS clase, count(*) AS cantidad 
            FROM {esquema}.{tabla}
            WHERE {filtro}
            GROUP BY {atributo_objetivo} ORDER BY cantidad DESC;
        """
        resultado = plpy.execute(query)

        cantidades = [fila["cantidad"] for fila in resultado]
        total = sum(cantidades)
        errores = sum(cantidades[1:])

        if Utils.eq(total, 0):
            return 0

        return errores + Estadistica.addErrores(total, errores, factor_confianza)

    def ampliar_filtro(self, filtro: str):
        return f"{filtro} AND {self.atributo} {self.comparador} '{self.atributo_valor}'"

    def errores_estimados(self):
        if self.tipo == "HOJA":
            return self.error_estimado(self.crear_filtro())
        else:
            total = 0
            for hijo in self.hijos:
                total += hijo.errores_estimados()
            return total

    def errores_estimados_rama(self, filtro: str):
        if self.tipo == "HOJA":
            return self.error_estimado(filtro)
        else:
            total = 0
            for hijo in self.hijos:
                total += hijo.errores_estimados_rama(hijo.ampliar_filtro(filtro))
            return total

    def podar(self) -> None:
        if self.tipo == "HOJA":
            return

        for hijo in self.hijos:
            hijo.podar()

        hijo_mas_grande = max(self.hijos, key=lambda hijo: hijo.cantidad_objetos)

        filtro = self.crear_filtro()

        if elevacion_subarbol:
            error_hijo_grande = hijo_mas_grande.errores_estimados_rama(filtro)
        else:
            # Reemplazar por el maximo de float despues
            error_hijo_grande = 99999999999999

        error_hoja = self.error_estimado(filtro)
        error_arbol = self.errores_estimados()

        if Utils.smOrEq(error_hoja, error_arbol + 0.1) and Utils.smOrEq(
            error_hoja, error_hijo_grande + 0.1
        ):
            self.dar_clase()
            return

        if Utils.smOrEq(error_hijo_grande, error_arbol + 0.1):
            self.hijos = hijo_mas_grande.hijos

            for hijo in self.hijos:
                hijo.padre = self

            if len(self.hijos) == 0:
                self.tipo = "HOJA"

            self.reevaluar()
            self.podar()

    def cantidad_nodos(self):
        cantidad = 1
        for hijo in self.hijos:
            cantidad += hijo.cantidad_nodos()

        return cantidad

    def cantidad_hojas(self):
        if self.tipo == "HOJA":
            cantidad = 1
        else:
            cantidad = 0
        for hijo in self.hijos:
            cantidad += hijo.cantidad_hojas()

        return cantidad

    def total_instancias_correctas(self):
        if self.tipo == "HOJA":
            cantidad = self.instancias_cantidad_correctas
        else:
            cantidad = 0
        for hijo in self.hijos:
            cantidad += hijo.total_instancias_correctas()

        return cantidad

    def total_instancias_incorrectas(self):
        if self.tipo == "HOJA":
            cantidad = self.instancias_cantidad_incorrectas
        else:
            cantidad = 0
        for hijo in self.hijos:
            cantidad += hijo.total_instancias_incorrectas()

        return cantidad

    def organizar_hijos(self):
        self.hijos = sorted(self.hijos, key=lambda hijo: len(hijo.hijos), reverse=True)

        for hijo in self.hijos:
            hijo.organizar_hijos()

    def obtener_cantidad_clase(self, cuenta: Dict[str, int]):
        if self.tipo == "HOJA" and self.clase is not None:
            cuenta[self.clase] += self.instancias_cantidad_correctas

        for hijo in self.hijos:
            hijo.obtener_cantidad_clase(cuenta)

    def obtener_clase_mas_probable(self):
        cantidad_clases: Dict[str, int] = defaultdict(int)
        self.obtener_cantidad_clase(cantidad_clases)
        return max(cantidad_clases, key=lambda k: cantidad_clases[k])

    def completar_atributos(self):
        if len(self.hijos) == 0:
            return

        for hijo in self.hijos:
            hijo.completar_atributos()

        atributo = self.hijos[0].atributo

        valores_hijos = [hijo.atributo_valor for hijo in self.hijos]
        valores_faltanes: List[str] = []

        if atributo not in Nodo.atributos_nominales_con_valores:
            return

        for valor in Nodo.atributos_nominales_con_valores[atributo]:
            if not valor in valores_hijos:
                valores_faltanes.append(valor)

        if len(valores_faltanes) == 0:
            return

        clase_mas_probable = self.obtener_clase_mas_probable()

        for valor in valores_faltanes:
            nodo = Nodo(self, "HOJA", atributo, "=", valor, [])
            nodo.clase = clase_mas_probable
            self.hijos.append(nodo)

    def obtener_id_nodos(self, lista: List[int]):
        lista.append(self.id)
        for hijo in self.hijos:
            hijo.obtener_id_nodos(lista)

    def __repr__(self) -> str:
        return f"<Nodo={self.id} tipo={self.tipo} atributo={self.atributo} clase='{self.atributo_valor}' total={self.cantidad_objetos} cantidad={self.instancias_cantidad_correctas}/{self.instancias_cantidad_incorrectas}>"


query = f"""
SELECT column_name as columna 
FROM information_schema.columns 
WHERE table_name = '{tabla}' 
AND table_schema = '{esquema}';
"""

resultado = plpy.execute(query)

atributos: List[str] = []

for fila in resultado:
    if fila["columna"] != atributo_objetivo:
        atributos.append(fila["columna"])

raiz = Nodo(None, "RAIZ", "-1", "=", "-1", atributos)
raiz.crear_hijos()


# obtener atributos usados en cada nodo
atributos_nominales_usados: set[str] = set()


def agregar_atributos_usados(nodo: Nodo):
    if nodo.tipo != "RAIZ" and not es_numerico(nodo.atributo):
        atributos_nominales_usados.add(nodo.atributo)

    for hijo in nodo.hijos:
        agregar_atributos_usados(hijo)


agregar_atributos_usados(raiz)

for atributo in atributos_nominales_usados:
    query = f"""
        SELECT DISTINCT {atributo} as atributo
        FROM {esquema}.{tabla}
    """
    resultado = plpy.execute(query)
    valores = [fila["atributo"] for fila in resultado]
    Nodo.atributos_nominales_con_valores[atributo] = valores

raiz.completar_atributos()

if colapsar_arbol:
    plpy.notice("Colapsando hijos que no disminuyan el error...")
    raiz.colapsar()

if raiz.cantidad_nodos() > 1:
    plpy.notice("Podando el arbol...")
    if factor_confianza >= 0.5:
        plpy.notice(
            "ADVERTENCIA: El factor de confianza es muy Alto (>0.5), Error estimado no modificado"
        )
    raiz.podar()

if raiz.cantidad_nodos() == 1:
    raiz.dar_clase()

raiz.organizar_hijos()

raiz.adicionar_en_tabla()
raiz.adicionar_hijos_en_tabla()

ids = []
raiz.obtener_id_nodos(ids)
Probabilidad.eliminar_probabilidades_sueltas(ids)
Probabilidad.adicionar_en_tabla()


def insertar_info(id, titulo, valor):
    plpy.execute(
        """
        INSERT INTO algoritmos_arbol.info (id, titulo, valor)
        VALUES ('%s', '%s', '%s')
    """
        % (id, titulo, valor)
    )


insertar_info("tipo", "Algoritmo Usado", "C45")
insertar_info("tabla", "Tabla", tabla)
insertar_info("attr_objetivo", "Atributo Objetivo", atributo_objetivo)
insertar_info("esquema", "Esquema", esquema)
insertar_info("num_nodos", "Número de Nodos", raiz.cantidad_nodos())
insertar_info("num_hojas", "Número de Hojas", raiz.cantidad_hojas())

return 'OK'
$$;
            
CREATE OR REPLACE FUNCTION
cross_validation(tabla VARCHAR, atributo_objetivo VARCHAR, esquema VARCHAR, algoritmo VARCHAR, parametros_adicionales VARCHAR, numero_folds INTEGER)
RETURNS VARCHAR
STRICT
LANGUAGE plpython3u
AS
$$
# tabla = "test"
# atributo_objetivo = "attr"
# esquema = "public"
# algoritmo = "c45"
# parametros_adicionales = ", 10, 0.25"
# numero_folds = 10

# Limpiar de llamados anteriores:
query = """
    TRUNCATE algoritmos_arbol.matriz_confusion;
    TRUNCATE algoritmos_arbol.clasificados CASCADE;
    TRUNCATE algoritmos_arbol.clasificados_probabilidades;
    """
plpy.execute(query)

plpy.notice("\n====== Crossfold Validation =======\n")

# obtener número de filas por fold
query = f"SELECT count(*) as cantidad FROM {esquema}.{tabla}"
total_filas: int = plpy.execute(query)[0]["cantidad"]
cantidad_iteracion: int = total_filas // numero_folds

restante: int = total_filas - (cantidad_iteracion * numero_folds)

nombre_tabla = f"{esquema}_{tabla}_entrenamiento"

query = f"""
    DROP TABLE IF EXISTS algoritmos_arbol_datos.{nombre_tabla}
"""
plpy.execute(query)

query = f"""
    CREATE TABLE algoritmos_arbol_datos.{nombre_tabla} AS
    SELECT * 
    FROM {esquema}.{tabla}
    ORDER BY random()
"""
plpy.execute(query)

# Añadir id temporal
query = f"""
    ALTER TABLE algoritmos_arbol_datos.{nombre_tabla}
    ADD COLUMN _id_temporal_pruebas INTEGER;
"""
plpy.execute(query)

# Darle valor al id temporal
query = f"""
    WITH numbered_rows AS (
      SELECT *, ROW_NUMBER() OVER () AS row_num, ctid
      FROM algoritmos_arbol_datos.{nombre_tabla}
    )
    UPDATE algoritmos_arbol_datos.{nombre_tabla} AS e
    SET _id_temporal_pruebas = nr.row_num
    FROM numbered_rows nr
    WHERE e.ctid = nr.ctid;
"""
plpy.execute(query)

# ====== Crear matriz de confusion con datos 0 ==========
query = f"""
    SELECT DISTINCT {atributo_objetivo} as clase
    FROM {esquema}.{tabla}
"""

resultado = plpy.execute(query)
clases = [fila["clase"] for fila in resultado]

valores_texto = []
for fila in clases:
    for columna in clases:
        valores_texto.append(f"('{fila}', '{columna}', 0)")

query = f"""
    INSERT INTO algoritmos_arbol.matriz_confusion
    (fila, columna, cantidad)
    VALUES
    {",".join(valores_texto)}
"""
plpy.execute(query)


#  =========== iterar sobre los folds =========
offset = 0

for i in range(numero_folds):
    plpy.notice(f"\n>>> Evaluando fold {i+1}/{numero_folds} <<<")

    query = "DROP TABLE IF EXISTS algoritmos_arbol_datos.fold_prueba"
    plpy.execute(query)
    query = "DROP TABLE IF EXISTS algoritmos_arbol_datos.fold_entrenamiento"
    plpy.execute(query)

    if restante > 0:
        limite = cantidad_iteracion + 1
    else:
        limite = cantidad_iteracion

    query = f"""
        CREATE TABLE algoritmos_arbol_datos.fold_prueba AS
        SELECT * 
        FROM algoritmos_arbol_datos.{nombre_tabla}
        OFFSET {offset}
        LIMIT {limite}
    """
    plpy.execute(query)

    offset += limite
    restante -= 1

    query = f"""
        CREATE TABLE algoritmos_arbol_datos.fold_entrenamiento AS
        SELECT * 
        FROM algoritmos_arbol_datos.{nombre_tabla}
        WHERE _id_temporal_pruebas NOT IN (
            SELECT _id_temporal_pruebas
            FROM algoritmos_arbol_datos.fold_prueba
        )
    """
    plpy.execute(query)

    # Eliminar columna de id temporal
    query = f"""
    ALTER TABLE algoritmos_arbol_datos.fold_prueba
    DROP COLUMN _id_temporal_pruebas
    """
    plpy.execute(query)

    query = f"""
    ALTER TABLE algoritmos_arbol_datos.fold_entrenamiento
    DROP COLUMN _id_temporal_pruebas
    """
    plpy.execute(query)

    # crear arbol
    query = f"""
        SELECT * FROM algoritmos_arbol.crear_arbol_{algoritmo}(
            'fold_entrenamiento',
            '{atributo_objetivo}',
            'algoritmos_arbol_datos'
            {parametros_adicionales}
        )
    """
    plpy.execute(query)

    # añadir a matriz y clasificados
    query = """
        SELECT * 
        FROM algoritmos_arbol.evaluar_matriz(
            'fold_prueba',
            'algoritmos_arbol_datos',
            TRUE
        )
    """
    resultado = plpy.execute(query)


# Insertar como serían las predicciones sin el arbol, luego crear las estadisticas de evaluación

query = f"""
    TRUNCATE TABLE algoritmos_arbol.cantidades_anteriores;
    
    INSERT INTO algoritmos_arbol.cantidades_anteriores
    (
        SELECT
            {atributo_objetivo} as clase,
            count(*) as cantidad
        FROM {esquema}.{tabla}
        GROUP BY {atributo_objetivo}
    );

    SELECT * FROM algoritmos_arbol.crear_estadisticas();
"""
plpy.execute(query)


plpy.notice(f"\n>>> Creando arbol con todos los datos <<<")
# arbol con todos los datos
query = f"""
    SELECT * FROM algoritmos_arbol.crear_arbol_{algoritmo}(
        '{tabla}',
        '{atributo_objetivo}',
        '{esquema}'
        {parametros_adicionales}
    )
"""
plpy.execute(query)


# Limpieza
plpy.execute(f"DROP TABLE IF EXISTS algoritmos_arbol_datos.{nombre_tabla}")
plpy.execute("DROP TABLE IF EXISTS algoritmos_arbol_datos.fold_prueba")
plpy.execute("DROP TABLE IF EXISTS algoritmos_arbol_datos.fold_entrenamiento")

return 'OK'
$$;
            
CREATE OR REPLACE FUNCTION
mostrar_arbol()
RETURNS VARCHAR
STRICT
LANGUAGE plpython3u
AS
$$
def construir_arbol_desde_tabla():
    def obtener_hijos(id_padre):
        query = f"""
        SELECT * FROM algoritmos_arbol.nodos 
        WHERE id_padre = {id_padre} 
        """
        return plpy.execute(query)

    def construir_nodo(nodo, es_ultimo=False, prefijo=""):
        arbol_str = ""

        if nodo["id_padre"] == -1:
            query = "SELECT valor from algoritmos_arbol.info WHERE id = 'tabla'"
            tabla = plpy.execute(query)[0]["valor"]
            arbol_str += f"· └─ {tabla}"
            query = "SELECT count(*) as cuenta from algoritmos_arbol.nodos"
            cuenta = plpy.execute(query)[0]["cuenta"]
            if cuenta == 1:
                total = int(
                    nodo["instancias_cantidad_correctas"]
                    + nodo["instancias_cantidad_incorrectas"]
                )
                arbol_str += f" -> {nodo['clase']} - ({total} / {nodo['instancias_cantidad_incorrectas']}) - {int(nodo['instancias_cantidad_correctas'])*100/total:.2f}%\n"
            nuevo_prefijo = "·    "
        else:
            if es_ultimo:
                arbol_str += f"{prefijo}└─ "
                nuevo_prefijo = prefijo + "   "
            else:
                arbol_str += f"{prefijo}├─ "
                nuevo_prefijo = prefijo + "│  "

            arbol_str += (
                f"{nodo['atributo']} {nodo['comparador']} {nodo['atributo_valor']}"
            )

        if nodo["tipo"] == "HOJA":
            if nodo["instancias_cantidad_incorrectas"] == 0:
                arbol_str += (
                    f" -> {nodo['clase']} - ({nodo['instancias_cantidad_correctas']}) - 100.00%\n"
                )
            else:
                total = int(
                    nodo["instancias_cantidad_correctas"]
                    + nodo["instancias_cantidad_incorrectas"]
                )
                arbol_str += f" -> {nodo['clase']} - ({total} / {nodo['instancias_cantidad_incorrectas']}) - {int(nodo['instancias_cantidad_correctas'])*100/total:.2f}%\n"
        else:
            arbol_str += "\n"
            hijos = obtener_hijos(nodo["id"])
            for i, hijo in enumerate(hijos):
                arbol_str += construir_nodo(hijo, i == len(hijos) - 1, nuevo_prefijo)

        return arbol_str

    raiz = plpy.execute("SELECT * FROM algoritmos_arbol.nodos WHERE id_padre = -1")[0]

    return construir_nodo(raiz)



arbol_str = "==== Arbol ====\n"
arbol_str += construir_arbol_desde_tabla()

hojas = plpy.execute("SELECT valor FROM algoritmos_arbol.info WHERE id = 'num_hojas'")[0]["valor"]
nodos = plpy.execute("SELECT valor FROM algoritmos_arbol.info WHERE id = 'num_nodos'")[0]["valor"]

arbol_str += "\n"
#arbol_str += f"\nNúmero de hojas: {hojas}\n"
#arbol_str += f"\nTamaño del arbol: {nodos}\n"

 
plpy.execute("delete from algoritmos_arbol.textos_evaluacion where id='arbol'")
plpy.execute("insert into algoritmos_arbol.textos_evaluacion(id, valor) values ('arbol', '{}')".format(arbol_str.replace("'", "''")))


plpy.notice(arbol_str)

return 'OK'
$$;
            
CREATE OR REPLACE FUNCTION
mostrar_info_arbol()
RETURNS VARCHAR
STRICT
LANGUAGE plpython3u
AS
$$
query = """
    SELECT length(titulo) as longitud
    FROM algoritmos_arbol.info
    ORDER BY length(titulo) DESC
    LIMIT 1;
"""
espacio_izquierdo = plpy.execute(query)[0]["longitud"]
espacio_izquierdo += 2

query = """
SELECT titulo, valor
FROM algoritmos_arbol.info
"""

resultado = plpy.execute(query)

info_str = "==== Información del Arbol ====";
for fila in resultado:
    info_str += f"\n{fila["titulo"]:<{espacio_izquierdo}} {fila["valor"]}"

info_str += "\n\n"


plpy.execute("delete from algoritmos_arbol.textos_evaluacion where id='info_arbol'")
plpy.execute("insert into algoritmos_arbol.textos_evaluacion(id, valor) values ('info_arbol', '{}')".format(info_str.replace("'", "''")))

plpy.notice(info_str)

return 'OK'
$$;
            
CREATE OR REPLACE FUNCTION
mostrar_evaluacion()
RETURNS VARCHAR
STRICT
LANGUAGE plpython3u
AS
$$
from collections import defaultdict
from typing import Dict

# Crear tabla
headers = {
    "Clase": "clase",
    "TPR": "tpr",
    "FPR": "fpr",
    "Precision": "presicion",
    "Recall": "recall",
    "F-Measure": "f_measure",
    "MCC": "mcc",
    "ROC-Area": "roc_area",
    "PRC-Area": "prc_area",
}
table_data = []

query = f"""
    SELECT tipo, {", ".join(headers.values())}
    FROM algoritmos_arbol.estadisticas_evaluacion
"""

resultado = plpy.execute(query)

for fila in resultado:
    row = []

    for columna in headers.values():
        value = fila[columna]

        if value is None:
            row.append("-")
        elif isinstance(value, float):
            row.append(f"{value:.3f}")
        else:
            row.append(str(value))
    table_data.append(row)

max_widths = []
for i in range(len(headers)):
    max_length = 0
    for row in table_data + [list(headers.values())]:
        length = len(str(row[i]))
        if length > max_length:
            max_length = length
    max_widths.append(max_length)

table_str = "==== Presición detallada por clase ====\n"

table_str += (
    "  ".join(f"{header:<{max_widths[i]}}" for i, header in enumerate(headers)) + "\n"
)

for row in table_data:
    table_str += (
        "  ".join(f"{str(item):<{max_widths[i]}}" for i, item in enumerate(row)) + "\n"
    )

table_str += "\n"

plpy.execute("delete from algoritmos_arbol.textos_evaluacion where id='presicion_clases'")
plpy.execute("insert into algoritmos_arbol.textos_evaluacion(id, valor) values ('presicion_clases', '{}')".format(table_str))

plpy.notice(table_str)


# string para mostrar la matriz
def nombre_columna(indice: int) -> str:
    indice += 1
    resultado = ""
    while indice > 0:
        indice -= 1
        resultado = chr(97 + (indice % 26)) + resultado
        indice //= 26
    return resultado


matriz: Dict[str, Dict[str, int]] = defaultdict(dict)

# Crear matriz
query = """
    SELECT fila, columna, cantidad
    FROM algoritmos_arbol.matriz_confusion
    ORDER BY fila, columna
"""

resultado = plpy.execute(query)

for fila_resultado in resultado:
    fila: str = fila_resultado["fila"]
    columna: str = fila_resultado["columna"]
    cantidad: int = fila_resultado["cantidad"]

    matriz[fila][columna] = cantidad

cantidad_maxima = max(max(matriz_filas.values()) for matriz_filas in matriz.values())
longitud = len(str(cantidad_maxima)) + 1

matriz_str = "==== Matriz de Confusión ====\n"
for indice in range(len(matriz)):
    matriz_str += nombre_columna(indice).ljust(longitud)

matriz_str += "  <-- Clasificado como"

for indice, (clase, clase_clasificada) in enumerate(matriz.items()):
    matriz_str += "\n"
    for cantidad in clase_clasificada.values():
        cantidad_str = str(cantidad)
        matriz_str += cantidad_str.ljust(longitud)
    matriz_str += "| "
    matriz_str += nombre_columna(indice)
    matriz_str += " = "
    matriz_str += clase

matriz_str += "\n"

plpy.execute("delete from algoritmos_arbol.textos_evaluacion where id='matriz'")
plpy.execute("insert into algoritmos_arbol.textos_evaluacion(id, valor) values ('matriz', '{}')".format(matriz_str.replace("'", "''")))

plpy.notice(matriz_str)

return 'OK'
$$;
            
CREATE OR REPLACE FUNCTION
mostrar_info_evaluacion()
RETURNS VARCHAR
STRICT
LANGUAGE plpython3u
AS
$$
query = """
    SELECT length(titulo) as longitud
    FROM algoritmos_arbol.info_evaluacion
    ORDER BY length(titulo) DESC
    LIMIT 1;
"""
espacio_izquierdo = plpy.execute(query)[0]["longitud"]
espacio_izquierdo += 2

texto_a_mostrar = "==== Información de la Evaluación ====\n"

query = """
    SELECT titulo, valor
    FROM algoritmos_arbol.info_evaluacion
    WHERE id = 'inst_total'
"""
resultado = plpy.execute(query)[0]
total = int(resultado["valor"])
texto_a_mostrar += f"{resultado["titulo"]:<{espacio_izquierdo}} {total}\n"

query = """
    SELECT id, titulo, valor
    FROM algoritmos_arbol.info_evaluacion
    WHERE id = 'inst_correctas'
"""
resultado = plpy.execute(query)[0]
cantidad_correctas = int(resultado["valor"])
porcentaje = 100 * cantidad_correctas / total
texto_a_mostrar += f"{resultado["titulo"]:<{espacio_izquierdo}} {cantidad_correctas}   {porcentaje:.2f}%\n"

query = """
    SELECT id, titulo, valor
    FROM algoritmos_arbol.info_evaluacion
    WHERE id = 'inst_incorrectas'
"""
resultado = plpy.execute(query)[0]
cantidad_incorrectas = int(resultado["valor"])
porcentaje = 100 * cantidad_incorrectas / total
texto_a_mostrar += f"{resultado["titulo"]:<{espacio_izquierdo}} {cantidad_incorrectas}   {porcentaje:.2f}%\n"

query = """
    SELECT id, titulo, valor
    FROM algoritmos_arbol.info_evaluacion
    WHERE id = 'kappa'
"""
resultado = plpy.execute(query)[0]
valor = resultado["valor"]
texto_a_mostrar += f"{resultado["titulo"]:<{espacio_izquierdo}} {valor:.4f}\n"

query = """
    SELECT id, titulo, valor
    FROM algoritmos_arbol.info_evaluacion
    WHERE id = 'err_abs_medio'
"""
resultado = plpy.execute(query)[0]
valor = resultado["valor"]
texto_a_mostrar += f"{resultado["titulo"]:<{espacio_izquierdo}} {valor:.4f}\n"

query = """
    SELECT id, titulo, valor
    FROM algoritmos_arbol.info_evaluacion
    WHERE id = 'raiz_err_abs_cuad'
"""
resultado = plpy.execute(query)[0]
valor = resultado["valor"]
texto_a_mostrar += f"{resultado["titulo"]:<{espacio_izquierdo}} {valor:.4f}\n"

query = """
    SELECT id, titulo, valor
    FROM algoritmos_arbol.info_evaluacion
    WHERE id = 'err_abs_rel'
"""
resultado = plpy.execute(query)[0]
valor = resultado["valor"]
texto_a_mostrar += f"{resultado["titulo"]:<{espacio_izquierdo}} {valor:.2f}%\n"

query = """
    SELECT id, titulo, valor
    FROM algoritmos_arbol.info_evaluacion
    WHERE id = 'raiz_err_cuad_rel'
"""
resultado = plpy.execute(query)[0]
valor = resultado["valor"]
texto_a_mostrar += f"{resultado["titulo"]:<{espacio_izquierdo}} {valor:.2f}%\n\n"


plpy.execute("delete from algoritmos_arbol.textos_evaluacion where id='info_evaluacion'")
plpy.execute("insert into algoritmos_arbol.textos_evaluacion(id, valor) values ('info_evaluacion', '{}')".format(texto_a_mostrar.replace("'", "''")))

plpy.notice(texto_a_mostrar)

return 'OK'
$$;
            
CREATE OR REPLACE FUNCTION
evaluar_matriz(tabla VARCHAR, esquema VARCHAR, acumular_valores BOOLEAN)
RETURNS VARCHAR
STRICT
LANGUAGE plpython3u
AS
$$
# tabla = "test"
# esquema = "public"
# acumular_valores = False

if not acumular_valores:
    query = """
        TRUNCATE algoritmos_arbol.matriz_confusion;
        TRUNCATE algoritmos_arbol.clasificados CASCADE;
        TRUNCATE algoritmos_arbol.clasificados_probabilidades;
    """
    plpy.execute(query)

from typing import Dict, List, Literal

class Filtro:
    clase: str
    filtro: str
    probabilidades: Dict[str, float]

    def __init__(self, clase: str, filtro: str, probabilidades: Dict[str, float]):
        self.clase = clase
        self.filtro = filtro
        self.probabilidades = probabilidades

    def __str__(self) -> str:
        return f"""
            <Filtro
                clase: {self.clase}   
                filtro: {self.filtro}   
                probabilidades: {self.probabilidades}   
            >
        """


filtros: List[Filtro] = []

def crear_filtros(id_nodo: int, condicion: str):
    query = f"""
        SELECT tipo, atributo, comparador, atributo_valor
        FROM algoritmos_arbol.nodos
        WHERE id = {id_nodo};
    """
    resultado = plpy.execute(query)[0]

    atributo = resultado["atributo"]
    comparador = resultado["comparador"]
    atributo_valor = resultado["atributo_valor"]

    condicion = f"{condicion} AND {atributo} {comparador} '{atributo_valor}'"

    tipo: Literal["RAIZ", "RAMA", "HOJA"] = resultado["tipo"]

    if tipo == "HOJA":
        query = f"""
            SELECT 
                clase,
                instancias_cantidad_correctas,
                instancias_cantidad_incorrectas
            FROM algoritmos_arbol.nodos
            WHERE id = {id_nodo};
        """
        clase = plpy.execute(query)[0]["clase"]

        query = f"""
            SELECT clase, probabilidad
            FROM algoritmos_arbol.clases_probabilidades
            WHERE nodo = {id_nodo}
        """

        resultado = plpy.execute(query)

        probabilidades: Dict[str, float] = {
            fila["clase"]: fila["probabilidad"] for fila in resultado
        }

        filtro = Filtro(clase, condicion, probabilidades)

        filtros.append(filtro)

        return

    query = f"""
        SELECT id
        FROM algoritmos_arbol.nodos
        WHERE id_padre = {id_nodo};
    """
    resultado = plpy.execute(query)

    for fila in resultado:
        crear_filtros(fila["id"], condicion)


query = """
    SELECT valor as atributo_objetivo
    FROM algoritmos_arbol.info
    WHERE id = 'attr_objetivo';
    """
resultado = plpy.execute(query)[0]
atributo_objetivo: str = resultado["atributo_objetivo"]


query = f"""
    SELECT DISTINCT {atributo_objetivo} AS clase
    FROM {esquema}.{tabla}
    UNION
    SELECT DISTINCT clase
    FROM algoritmos_arbol.nodos
    WHERE clase IS NOT NULL
    """
resultado = plpy.execute(query)
matriz: Dict[str, Dict[str, int]] = {}

# Crear matriz
for fila in resultado:
    clases_matriz = {}
    for fila_interno in resultado:
        clases_matriz[fila_interno["clase"]] = 0

    matriz[fila["clase"]] = clases_matriz


query = f"""
    SELECT id
    FROM algoritmos_arbol.nodos
    WHERE id_padre = 0;
"""
resultado = plpy.execute(query)

if len(resultado) == 0:
    query = f"""
        SELECT clase
        FROM algoritmos_arbol.nodos
        WHERE tipo = 'RAIZ'
    """
    resultado = plpy.execute(query)[0]

    clase = resultado["clase"]

    query = f"""
        SELECT clase, probabilidad
        FROM algoritmos_arbol.clases_probabilidades
        WHERE nodo = 0
    """

    resultado = plpy.execute(query)

    probabilidades: Dict[str, float] = {
        fila["clase"]: fila["probabilidad"] for fila in resultado
    }

    filtro = Filtro(clase, "TRUE", probabilidades)

    filtros.append(filtro)
else:
    for fila in resultado:
        crear_filtros(fila["id"], "true")



for filtro in filtros:
    query = f"""
        SELECT
            count(*) as cuenta,
            {atributo_objetivo} as clase
        FROM {esquema}.{tabla}
        WHERE {filtro.filtro}
        GROUP BY {atributo_objetivo}
    """
    resultado = plpy.execute(query)

    for fila in resultado:
        query = f"""
            INSERT INTO algoritmos_arbol.clasificados
            (clase, cantidad)
            VALUES ('{fila["clase"]}', {fila["cuenta"]})
            RETURNING id;
        """
        id = plpy.execute(query)[0]["id"]

        probabilidades_texto: List[str] = []
        for clase, valor in filtro.probabilidades.items():
            probabilidades_texto.append(f"({id}, '{clase}', {valor})")


        query = f"""
            INSERT INTO algoritmos_arbol.clasificados_probabilidades
            (id, clase, probabilidad)
            VALUES
            {",".join(probabilidades_texto)}
        """

        plpy.execute(query)

        clases_matriz = matriz[fila["clase"]]
        clases_matriz[filtro.clase] += fila["cuenta"]


for fila, clase_clasificada in matriz.items():
    for columna, cantidad in clase_clasificada.items():
        if acumular_valores:
            query = f"""
                UPDATE algoritmos_arbol.matriz_confusion
                    SET cantidad = cantidad + {cantidad}
                WHERE 
                    columna = '{columna}'
                    AND fila = '{fila}'
            """
        else:
            query = """
                INSERT INTO algoritmos_arbol.matriz_confusion
                ( fila, columna, cantidad)
                VALUES ('%s', '%s', %s)
            """ % (
                fila,
                columna,
                cantidad,
            )

        plpy.execute(query)

# Insertar como serían las predicciones sin el arbol
query = f"""
    TRUNCATE TABLE algoritmos_arbol.cantidades_anteriores;
    INSERT INTO algoritmos_arbol.cantidades_anteriores
    (
        SELECT
            {atributo_objetivo} as clase,
            count(*) as cantidad
        FROM {esquema}.{tabla}
        GROUP BY {atributo_objetivo}
    )
"""
plpy.execute(query)

return 'OK'
$$;
            
CREATE OR REPLACE FUNCTION
crear_estadisticas()
RETURNS VARCHAR
STRICT
LANGUAGE plpython3u
AS
$$
# tabla = "test"
# atributo_objetivo = "test2"
# esquema = "public"

from collections import defaultdict
from math import sqrt
from typing import Dict, List, Optional

query = "TRUNCATE TABLE algoritmos_arbol.estadisticas_evaluacion"
plpy.execute(query)


class ResultadoEstadistico:
    vp: int
    fp: int
    vn: int
    fn: int

    def __init__(self, vp: int, fp: int, vn: int, fn: int):
        self.vp = vp
        self.fp = fp
        self.vn = vn
        self.fn = fn

    @classmethod
    def desde_clase(cls, clase_buscada: str, matriz: Dict[str, Dict[str, int]]):
        vp = 0
        fp = 0
        vn = 0
        fn = 0

        for clase_fila, columnas in matriz.items():
            if clase_fila == clase_buscada:
                for columna, cantidad in columnas.items():

                    if columna == clase_buscada:
                        vp += cantidad
                    else:
                        fn += cantidad
            else:
                for columna, cantidad in columnas.items():
                    if columna == clase_buscada:
                        fp += cantidad
                    else:
                        vn += cantidad

        return cls(vp, fp, vn, fn)

    def total_clase(self):
        return self.vp + self.fn

    def tpr(self) -> Optional[float]:
        vp = self.vp
        fn = self.fn
        numerador = vp
        denominador = vp + fn
        if denominador == 0:
            return None
        return numerador / denominador

    def recall(self) -> Optional[float]:
        return self.tpr()  # Recall es igual a TPR

    def tnr(self) -> Optional[float]:
        fp = self.fp
        vn = self.vn
        numerador = vn
        denominador = vn + fp
        if denominador == 0:
            return None
        return numerador / denominador

    def fpr(self) -> Optional[float]:
        fp = self.fp
        vn = self.vn
        numerador = fp
        denominador = fp + vn
        if denominador == 0:
            return None
        return numerador / denominador

    def fnr(self) -> Optional[float]:
        vp = self.vp
        fn = self.fn
        numerador = fn
        denominador = fn + vp
        if denominador == 0:
            return None
        return numerador / denominador

    def presicion(self) -> Optional[float]:
        vp = self.vp
        fp = self.fp
        numerador = vp
        denominador = vp + fp
        if denominador == 0:
            return None
        return numerador / denominador

    def f_score(self) -> Optional[float]:
        recall = self.tpr()
        presicion = self.presicion()
        if recall is None or presicion is None:
            return None
        numerador = 2 * recall * presicion
        denominador = recall + presicion
        if denominador == 0:
            return None
        return numerador / denominador

    def mcc(self) -> Optional[float]:
        vp = self.vp
        fp = self.fp
        vn = self.vn
        fn = self.fn
        numerador = (vp * vn) - (fp * fn)
        denominador = sqrt((vp + fp) * (vp + fn) * (vn + fp) * (vn + fn))
        if denominador == 0:
            return None
        return numerador / denominador

    def __repr__(self) -> str:
        return f"<vp={self.vp} fp={self.fp} vn={self.vn} fn={self.fn}>"


class Prediccion:
    clase_actual: str
    probabilidades: Dict[str, float]
    cantidad: int

    predicciones: List["Prediccion"] = []

    def __init__(
        self,
        clase_actual: str,
        probabilidades: Dict[str, float],
        cantidad: int,
    ) -> None:
        self.clase_actual = clase_actual
        self.probabilidades = probabilidades
        self.cantidad = cantidad

    def __repr__(self) -> str:
        return f"<clase_actual={self.clase_actual} cantidad={self.cantidad} probs={self.probabilidades}>"

    @staticmethod
    def adicionar_prediccion(
        clase_actual: str, probabilidades: Dict[str, float], cantidad: int
    ):
        prediccion = Prediccion(clase_actual, probabilidades, cantidad)
        Prediccion.predicciones.append(prediccion)

    @staticmethod
    def obtener_ordenado(clase: str) -> List["Prediccion"]:
        lista_ordenada = sorted(
            Prediccion.predicciones, key=lambda pred: pred.probabilidades[clase]
        )
        return lista_ordenada


class Instancia:
    verdaderos_positivos: int
    falsos_positivos: int
    recall: Optional[float]
    presicion: Optional[float]

    def __init__(
        self,
        positivos_verdaderos: int,
        positivos_falsos: int,
        recall: Optional[float],
        presicion: Optional[float],
    ) -> None:
        self.verdaderos_positivos = positivos_verdaderos
        self.falsos_positivos = positivos_falsos
        self.recall = recall
        self.presicion = presicion

    def __repr__(self) -> str:
        recall = f"{self.recall:.3f}" if self.recall is not None else "?"
        presicion = f"{self.presicion:.3f}" if self.presicion is not None else "?"
        return f"<tp={self.verdaderos_positivos} fp={self.falsos_positivos} recall={recall} presicion={presicion}>"


class UmbralCurva:
    predicciones: List[Prediccion]
    instancias: List[Instancia]
    clase: str

    def __init__(self, clase: str) -> None:
        self.predicciones = Prediccion.obtener_ordenado(clase)
        self.clase = clase
        self.crear_curva()

    def crear_curva(self) -> None:
        self.instancias = []

        # Obtener total positivos y total negativos
        query = f"""
            SELECT clase, cantidad
            FROM algoritmos_arbol.cantidades_anteriores
        """
        resultado = plpy.execute(query)

        total_positivos = 0
        total_negativos = 0

        for fila in resultado:
            if fila["clase"] == self.clase:
                total_positivos += fila["cantidad"]
            else:
                total_negativos += fila["cantidad"]

        tc = ResultadoEstadistico(total_positivos, total_negativos, 0, 0)

        positivos = 0
        negativos = 0

        i = 0

        while True:
            tc.vp = tc.vp - positivos
            tc.fn = tc.fn + positivos
            tc.fp = tc.fp - negativos
            tc.vn = tc.vn + negativos

            instancia = Instancia(tc.vp, tc.fp, tc.recall(), tc.presicion())
            self.instancias.append(instancia)

            positivos = 0
            negativos = 0

            iguales = True
            while iguales:
                prediccion = self.predicciones[i]
                if prediccion.clase_actual == self.clase:
                    positivos += prediccion.cantidad
                else:
                    negativos += prediccion.cantidad

                if i == len(self.predicciones) - 1 or (
                    self.predicciones[i + 1].probabilidades[clase]
                    != prediccion.probabilidades[clase]
                ):
                    iguales = False

                i += 1

            if i == len(self.predicciones):
                break

        if tc.fn != total_positivos or tc.vn != total_negativos:
            tc = ResultadoEstadistico(0, 0, total_negativos, total_positivos)
            instancia = Instancia(tc.vp, tc.fp, tc.recall(), tc.presicion())
            self.instancias.append(instancia)

    def area_roc(self) -> Optional[float]:
        n = len(self.instancias)

        if n == 0:
            return None

        area = 0
        cum_neg = 0

        instancias = self.instancias

        total_pos = instancias[0].verdaderos_positivos
        total_neg = instancias[0].falsos_positivos

        for i in range(n):
            if i < n - 1:
                cip = (
                    instancias[i].verdaderos_positivos
                    - instancias[i + 1].verdaderos_positivos
                )
                cin = (
                    instancias[i].falsos_positivos - instancias[i + 1].falsos_positivos
                )
            else:
                cip = instancias[n - 1].verdaderos_positivos
                cin = instancias[n - 1].falsos_positivos

            area += cip * (cum_neg + (0.5 * cin))
            cum_neg += cin

        if total_neg * total_pos == 0:
            return None

        area /= total_neg * total_pos

        return area

    def area_prc(self) -> Optional[float]:
        n = len(self.instancias)

        if n == 0:
            return None

        area = 0

        instancias = self.instancias

        xlast = instancias[n - 1].recall

        for i in range(n - 2, -1, -1):
            recall = instancias[i].recall
            if recall is None or xlast is None:
                return None

            recall_delta = recall - xlast

            presicion = instancias[i].presicion
            if presicion is None or xlast is None:
                return None

            area += presicion * recall_delta

            xlast = instancias[i].recall

        if area == 0:
            return None

        return area

    def __repr__(self) -> str:
        texto = f"predicciones: {self.predicciones}\n"
        texto += f"instancias: {self.instancias}\n"
        texto += f"clase: {self.clase}\n"

        return texto


class Estadisticas:
    suma_errores: float
    suma_errores_abs: float
    suma_errores_sqr: float
    suma_errores_anterior_abs: float
    suma_errores_anterior_sqr: float

    total_matriz: int
    total_tabla: int

    total_correctas: int
    total_incorrectas: int

    def __init__(self, matriz: Dict[str, Dict[str, int]]) -> None:
        self.actualizar_errores(matriz)

    def actualizar_errores(self, matriz: Dict[str, Dict[str, int]]):
        self.suma_errores = 0
        self.suma_errores_abs = 0
        self.suma_errores_sqr = 0
        self.suma_errores_anterior_abs = 0
        self.suma_errores_anterior_sqr = 0

        total_correctas = 0
        total_incorrectas = 0

        for fila, dict_col in matriz.items():
            for columna, valor in dict_col.items():
                if fila == columna:
                    total_correctas += valor
                else:
                    total_incorrectas += valor

        self.total_correctas = total_correctas
        self.total_incorrectas = total_incorrectas

        query = f"""
            SELECT clase, cantidad
            FROM algoritmos_arbol.cantidades_anteriores
        """

        resultado = plpy.execute(query)
        total: int = sum(fila["cantidad"] for fila in resultado)

        probabilidades_anteriores: dict[str, float] = defaultdict(float)
        for fila in resultado:
            probabilidades_anteriores[fila["clase"]] = fila["cantidad"] / total

        self.total_matriz = total_correctas + total_incorrectas
        self.total_tabla = total

        for prediccion in Prediccion.predicciones:
            suma_errores: float = 0
            suma_errores_abs: float = 0
            suma_errores_sqr: float = 0
            suma_errores_anterior_abs: float = 0
            suma_errores_anterior_sqr: float = 0

            for clase, probabilidad in prediccion.probabilidades.items():
                if clase == prediccion.clase_actual:
                    probabilidad_actual = 1
                else:
                    probabilidad_actual = 0

                diferencia = probabilidad - probabilidad_actual
                suma_errores += diferencia
                suma_errores_abs += abs(diferencia)
                suma_errores_sqr += diferencia**2

                diferencia = probabilidades_anteriores[clase] - probabilidad_actual
                suma_errores_anterior_abs += abs(diferencia)
                suma_errores_anterior_sqr += diferencia**2

            cantidad_probs = len(prediccion.probabilidades)

            self.suma_errores += suma_errores * prediccion.cantidad / cantidad_probs
            self.suma_errores_abs += (
                suma_errores_abs * prediccion.cantidad / cantidad_probs
            )
            self.suma_errores_sqr += (
                suma_errores_sqr * prediccion.cantidad / cantidad_probs
            )
            self.suma_errores_anterior_abs += (
                suma_errores_anterior_abs * prediccion.cantidad / cantidad_probs
            )
            self.suma_errores_anterior_sqr += (
                suma_errores_anterior_sqr * prediccion.cantidad / cantidad_probs
            )

    @staticmethod
    def kappa(matriz: Dict[str, Dict[str, int]]) -> float:
        etiquetas = list(matriz.keys())

        suma_filas = {clase: 0 for clase in etiquetas}
        suma_columnas = {clase: 0 for clase in etiquetas}
        suma_pesos = 0

        for fila in etiquetas:
            for columna in etiquetas:
                suma_filas[fila] += matriz[fila][columna]
                suma_columnas[columna] += matriz[fila][columna]
                suma_pesos += matriz[fila][columna]

        correcto = 0
        acuerdo_azar = 0

        for etiqueta in etiquetas:
            acuerdo_azar += suma_filas[etiqueta] * suma_columnas[etiqueta]
            correcto += matriz[etiqueta][etiqueta]

        acuerdo_azar /= suma_pesos * suma_pesos
        correcto /= suma_pesos

        if acuerdo_azar < 1:
            return (correcto - acuerdo_azar) / (1 - acuerdo_azar)
        else:
            return 1

    def error_absoluto_medio(self) -> float:
        return self.suma_errores_abs / self.total_matriz

    def error_absoluto_medio_anterior(self) -> float:
        return self.suma_errores_anterior_abs / self.total_tabla

    def raiz_error_absoluto_medio_cuadratico(self) -> float:
        return sqrt(self.suma_errores_sqr / self.total_matriz)

    def raiz_error_absoluto_medio_cuadratico_anterior(self) -> float:
        return sqrt(self.suma_errores_anterior_sqr / self.total_tabla)

    def error_absoluto_relativo(self) -> float:
        error = self.error_absoluto_medio_anterior()
        if error == 0:
            return 0
        return 100 * self.error_absoluto_medio() / self.error_absoluto_medio_anterior()

    def raiz_error_cuadratico_relativo(self) -> float:
        error = self.raiz_error_absoluto_medio_cuadratico_anterior()
        if error == 0:
            return 0

        return 100 * self.raiz_error_absoluto_medio_cuadratico() / error


# ========== Obtener matriz de confusion =========

query = """
    SELECT fila, columna, cantidad
    FROM algoritmos_arbol.matriz_confusion
"""

resultado = plpy.execute(query)

matriz: Dict[str, Dict[str, int]] = defaultdict(dict)
total_clases = 0

for fila_resultado in resultado:
    fila: str = fila_resultado["fila"]
    columna: str = fila_resultado["columna"]
    cantidad: int = fila_resultado["cantidad"]

    matriz[fila][columna] = cantidad
    total_clases += cantidad

# ========== Obtener Predicciones ===========
query = """
    SELECT id, clase, cantidad
    FROM algoritmos_arbol.clasificados
"""
resultado = plpy.execute(query)

for fila_resultado in resultado:
    query = f"""
        SELECT clase, probabilidad
        FROM algoritmos_arbol.clasificados_probabilidades
        WHERE id = {fila_resultado["id"]}
    """
    resultado_probabilidades = plpy.execute(query)

    probabilidades: Dict[str, float] = {
        fila["clase"]: fila["probabilidad"] for fila in resultado_probabilidades
    }

    Prediccion.adicionar_prediccion(
        fila_resultado["clase"], probabilidades, fila_resultado["cantidad"]
    )


# ======== Hacer los demas calculos =======


def actualizar_total_ponderado(
    total: Optional[float], valor: Optional[float], porcentaje: float
) -> Optional[float]:
    return (
        total + valor * porcentaje
        if isinstance(total, float) and isinstance(valor, float)
        else None
    )


totales_ponderados: Dict[str, float | None] = {
    "TPR": 0.0,
    "FPR": 0.0,
    "Precision": 0.0,
    "Recall": 0.0,
    "F-Measure": 0.0,
    "MCC": 0.0,
    "ROC-Area": 0.0,
    "PRC-Area": 0.0,
}


def verificar_entrada(valor: float | None):
    if valor is None:
        return "NULL"
    else:
        return valor


def insertar_en_tabla(diccionario, tipo: str, clase: str):
    query = """
        INSERT INTO algoritmos_arbol.estadisticas_evaluacion
        (
            tipo,
            clase,
            tpr,
            fpr,
            presicion,
            recall,
            f_measure,
            mcc,
            roc_area,
            prc_area
        )
        VALUES ('%s', '%s', %s, %s, %s, %s, %s, %s, %s, %s)
    """ % (
        tipo,
        clase,
        verificar_entrada(diccionario["TPR"]),
        verificar_entrada(diccionario["FPR"]),
        verificar_entrada(diccionario["Precision"]),
        verificar_entrada(diccionario["Recall"]),
        verificar_entrada(diccionario["F-Measure"]),
        verificar_entrada(diccionario["MCC"]),
        verificar_entrada(diccionario["ROC-Area"]),
        verificar_entrada(diccionario["PRC-Area"]),
    )

    plpy.execute(query)


for clase in matriz.keys():
    resultado = ResultadoEstadistico.desde_clase(clase, matriz)
    porcentaje = resultado.total_clase() / total_clases

    curva = UmbralCurva(clase)

    metricas = {
        "TPR": resultado.tpr(),
        "FPR": resultado.fpr(),
        "Precision": resultado.presicion(),
        "Recall": resultado.recall(),
        "F-Measure": resultado.f_score(),
        "MCC": resultado.mcc(),
        "ROC-Area": curva.area_roc(),
        "PRC-Area": curva.area_prc(),
    }

    for nombre, valor in metricas.items():
        totales_ponderados[nombre] = actualizar_total_ponderado(
            totales_ponderados[nombre], valor, porcentaje
        )

    # Añadir a base de datos
    insertar_en_tabla(metricas, "PARCIAL", clase)


insertar_en_tabla(totales_ponderados, "TOTAL", "Total Ponderado")


# ========== Estadisticas Generales ==========

# Eliminar estadisticas anteriores
plpy.execute("TRUNCATE TABLE algoritmos_arbol.info_evaluacion")


def insertar_info(id: str, titulo: str, valor: float):
    plpy.execute(
        """
        INSERT INTO algoritmos_arbol.info_evaluacion(id, titulo, valor)
        VALUES ('%s', '%s', %s)
    """
        % (id, titulo, valor)
    )


estadisticas = Estadisticas(matriz)
insertar_info(
    "inst_correctas",
    "Instancias Clasificadas Correctamente",
    estadisticas.total_correctas,
)
insertar_info(
    "inst_incorrectas",
    "Instancias Clasificadas Incorrectamente",
    estadisticas.total_incorrectas,
)
insertar_info("kappa", "Estadistica Kappa", estadisticas.kappa(matriz))
insertar_info(
    "err_abs_medio", "Error Absoluto Medio", estadisticas.error_absoluto_medio()
)
insertar_info(
    "raiz_err_abs_cuad",
    "Raiz del Error Medio Cuadratico",
    estadisticas.raiz_error_absoluto_medio_cuadratico(),
)
insertar_info(
    "err_abs_rel", "Error Absoluto Relativo", estadisticas.error_absoluto_relativo()
)
insertar_info(
    "raiz_err_cuad_rel",
    "Raiz del Error Cuadratico Relativo",
    estadisticas.raiz_error_cuadratico_relativo(),
)
insertar_info("inst_total", "Total de Instancias", estadisticas.total_matriz)

return 'OK'
$$;
            
CREATE OR REPLACE FUNCTION
predecir_tabla(tabla VARCHAR, esquema VARCHAR, cantidad_instancias INTEGER)
RETURNS VARCHAR
STRICT
LANGUAGE plpython3u
AS
$$
# esquema = "public"
# tabla = "test"
# cantidad_instancias = 100
from typing import List, Literal

query = """
    SELECT valor as atributo_objetivo
    FROM algoritmos_arbol.info
    WHERE id = 'attr_objetivo';
    """
resultado = plpy.execute(query)[0]
atributo_objetivo: str = resultado["atributo_objetivo"]


class Condicion:
    atributo: str
    comparador: str
    atributo_valor: str

    def __init__(self, atributo: str, comparador: str, atributo_valor: str) -> None:
        self.atributo = atributo
        self.comparador = comparador
        self.atributo_valor = atributo_valor


class Filtro:
    condiciones: List[Condicion]
    clase: str
    certeza: float

    def __init__(self) -> None:
        self.condiciones = []
        self.clase = ""
        self.certeza = 0

    def obtener_condicion_sql(self) -> str:
        sql = "true"
        for condicion in self.condiciones:
            sql += f" and {condicion.atributo} {condicion.comparador} '{condicion.atributo_valor}'"

        return sql

    def copiar_filtro(self) -> "Filtro":
        nuevo = Filtro()
        nuevo.condiciones = self.condiciones[::]
        return nuevo


filtros: List[Filtro] = []


def crear_filtros(id_nodo: int, filtro: Filtro):
    query = f"""
        SELECT tipo, atributo, comparador, atributo_valor
        FROM algoritmos_arbol.nodos
        WHERE id = {id_nodo};
    """
    resultado = plpy.execute(query)[0]

    atributo = resultado["atributo"]
    comparador = resultado["comparador"]
    atributo_valor = resultado["atributo_valor"]

    condicion = Condicion(atributo, comparador, atributo_valor)

    filtro.condiciones.append(condicion)

    tipo: Literal["RAIZ", "RAMA", "HOJA"] = resultado["tipo"]

    if tipo == "HOJA":
        query = f"""
            SELECT clase,
            instancias_cantidad_correctas,
            instancias_cantidad_incorrectas
            FROM algoritmos_arbol.nodos
            WHERE id = {id_nodo};
        """
        resultado = plpy.execute(query)[0]
        ins_correctas = resultado["instancias_cantidad_correctas"]
        ins_incorrectas = resultado["instancias_cantidad_incorrectas"]

        filtro.clase = resultado["clase"]
        total = ins_correctas + ins_incorrectas

        if total == 0:
            filtro.certeza = 1
        else:
            filtro.certeza = ins_correctas / total

        filtros.append(filtro)

    query = f"""
        SELECT id
        FROM algoritmos_arbol.nodos
        WHERE id_padre = {id_nodo};
    """
    resultado = plpy.execute(query)

    for fila in resultado:
        crear_filtros(fila["id"], filtro.copiar_filtro())


query = f"""
    SELECT id
    FROM algoritmos_arbol.nodos
    WHERE id_padre = 0;
"""
resultado = plpy.execute(query)

if len(resultado) == 0:
    query = f"""
        SELECT clase
        FROM algoritmos_arbol.nodos
        WHERE tipo = 'RAIZ'
        instancias_cantidad_correctas,
        instancias_cantidad_incorrectas
        FROM algoritmos_arbol.nodos;
        """
    resultado = plpy.execute(query)[0]
    ins_correctas = resultado["instancias_cantidad_correctas"]
    ins_incorrectas = resultado["instancias_cantidad_incorrectas"]

    filtro = Filtro()
    filtro.clase = resultado["clase"]
    filtro.certeza = ins_correctas / (ins_correctas + ins_incorrectas)

    resultado = plpy.execute(query)
    filtros.append(filtro)
else:
    for fila in resultado:
        crear_filtros(fila["id"], Filtro())

# Tabla donde se guardan las predicciones
query = f"""
DROP TABLE IF EXISTS algoritmos_arbol_datos.predicciones;
"""
plpy.execute(query)

query = f"""
CREATE TABLE algoritmos_arbol_datos.predicciones
AS TABLE {esquema}.{tabla};
"""
plpy.execute(query)

query = f"""
ALTER TABLE algoritmos_arbol_datos.predicciones
ADD COLUMN _actual TEXT DEFAULT NULL,
ADD COLUMN _predicho TEXT DEFAULT NULL,
ADD COLUMN _error BOOL DEFAULT FALSE,
ADD COLUMN _prediccion FLOAT DEFAULT NULL;
"""
plpy.execute(query)

for filtro in filtros:
    query = f"""
    UPDATE algoritmos_arbol_datos.predicciones
    SET _actual = {atributo_objetivo},
        _predicho = '{filtro.clase}',
        _prediccion = {filtro.certeza}
    WHERE {filtro.obtener_condicion_sql()};
    """
    plpy.execute(query)

query = f"""
UPDATE algoritmos_arbol_datos.predicciones
SET _error = true
WHERE not _actual = _predicho;
"""
plpy.execute(query)

# Resumen

query = f"""
    SELECT count(*) AS cantidad
    FROM algoritmos_arbol_datos.predicciones
    WHERE _predicho is not NULL
"""
predichos = plpy.execute(query)[0]["cantidad"]

query = f"""
    SELECT count(*) AS cantidad
    FROM algoritmos_arbol_datos.predicciones
    WHERE _predicho is NULL
"""
no_predichos = plpy.execute(query)[0]["cantidad"]

query = f"""
    SELECT count(*) AS cantidad
    FROM algoritmos_arbol_datos.predicciones
    WHERE _actual = _predicho
    AND _actual is not NULL
    AND _predicho is not NULL
"""
predichos_correctamente = plpy.execute(query)[0]["cantidad"]

query = f"""
    SELECT count(*) AS cantidad
    FROM algoritmos_arbol_datos.predicciones
    WHERE _actual != _predicho
    AND _actual is not NULL
    AND _predicho is not NULL
"""
predichos_incorrectamente = plpy.execute(query)[0]["cantidad"]

texto = "==== Resumen de Predicciones ====\n"
texto += f"Instancias predichas: {predichos}\n"
texto += f"Instancias no predichas: {no_predichos}\n"
texto += f"Instancias predichas correctamente: {predichos_correctamente}\n"
texto += f"Instancias predichas incorrectamente: {predichos_incorrectamente}\n"
texto += "\n"

plpy.execute(
    "delete from algoritmos_arbol.textos_evaluacion where id='resumen_prediccion'"
)
plpy.execute(
    "insert into algoritmos_arbol.textos_evaluacion(id, valor) values ('arbol', '{}')".format(
        texto.replace("'", "''")
    )
)
plpy.notice(texto)


if cantidad_instancias != 0:
    if cantidad_instancias < 0:
        limite = ""
    else:
        limite = f"LIMIT {cantidad_instancias}"
    # Mostrar predicciones
    query = f"""
        SELECT DISTINCT {atributo_objetivo} AS clase
        FROM algoritmos_arbol_datos.predicciones
        UNION
        SELECT DISTINCT clase
        FROM algoritmos_arbol.nodos
        WHERE clase IS NOT NULL
        """
    resultado = plpy.execute(query)
    clases = [fila["clase"] for fila in resultado if fila["clase"] is not None]
    clase_mas_larga = max(clases, key=len)

    query = "SELECT count(*) as total FROM algoritmos_arbol_datos.predicciones"
    resultado = plpy.execute(query)

    pad_instancia = len(str(resultado[0]["total"]))
    pad_instancia = max(len("Instancia#"), pad_instancia)
    pad_actual = max(len("Actual"), len(clase_mas_larga))
    pad_predicho = max(len("Predicho"), len(clase_mas_larga))

    texto = "==== Predicciones ====\n"

    texto += f"{"Instancia#".ljust(pad_instancia)}  {"Actual".ljust(pad_actual)}  {"Predicho".ljust(pad_predicho)}  Error  Predicción\n"

    query = f"""
        SELECT _actual, _predicho, _error, _prediccion
        FROM algoritmos_arbol_datos.predicciones
        {limite}
    """

    cursor = plpy.cursor(query)

    inst = 1
    while True:
        filas = cursor.fetch(5000)
        if not filas:
            break

        resultado = ""
        for fila in filas:
            actual: str = fila["_actual"]
            if actual is None:
                actual = "?"

            predicho: str = fila["_predicho"]
            if predicho is None:
                predicho = "?"

            error: bool = fila["_error"]

            prediccion: float | str = fila["_prediccion"]
            if prediccion is None:
                prediccion = "?"

            resultado += f"{str(inst).ljust(pad_instancia)}  {actual.ljust(pad_actual)}  {predicho.ljust(pad_predicho)}  {str(error).ljust(5)}  {prediccion:.3f}\n"
            inst += 1

        texto += resultado

    texto += "\n"
    plpy.execute(
        "delete from algoritmos_arbol.textos_evaluacion where id='instancias_predichas'"
    )
    plpy.execute(
        "insert into algoritmos_arbol.textos_evaluacion(id, valor) values ('arbol', '{}')".format(
            texto.replace("'", "''")
        )
    )
    plpy.notice(texto)

return 'OK'
$$;
            