# Funciones Principales
Estas funciones son basicamente envoltorios de las otras funciones para crear y evaluar un arbol de desición con un solo llamado de funciones.

## c45
Esta función crea un arbol utilizando el algoritmo c4.5, luego evalua la efectividad del arbol utilizando la misma tabla de entrenamiento.

### c45(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, objetos_minimos INTEGER, factor_confianza FLOAT)

Parametros:

- `tabla`: El nombre de la tabla que se usará para crear el arbol y evaluarlo
- `atributo`: El nombre del atributo que se va evaluar
- `esquema`: El esquema al cual pertence la tabla que se usará para crear el arbol y evaluarlo
- `objetos_minimos`: El minimo número de objetos para no seguir dividiendo una rama en el arbol de desición
- `factor_confianza`: Factor de confianza, se debe especificar un valor entre 0 y 1, un valor mas pequeño llevará a una poda mas agresiva

Ejemplo de uso:

```sql
SELECT * FROM c45("resultado_juego", "jugar", "public", 2, 0.3)
```

Esta instrucción creará una arbol de desición con el algoritmo c4.5 utilizando la tabla `resultado_juego`, la cual se encuentra en el esquema `public`, con el fin de predecir la columna `jugar`, el minimo de objetos por rama es de `2`, y utiliza un factor de confianza de `0.3`.


### c45(tabla VARCHAR, atributo VARCHAR)

Identica a la función anterior, pero toma por defecto al esquema como `public`, al minimo de objetos como 1 y al factor de confianza como `0.25`.

Parametros:

- `tabla`: El nombre de la tabla que se usará para crear el arbol y evaluarlo
- `atributo`: El nombre del atributo que se va evaluar

Ejemplo de uso:

```sql
SELECT * FROM c45("pasajeros_titanic", "sobrevivio")
```

Esta instrucción creará una arbol de desición con el algoritmo c4.5 utilizando la tabla `pasajeros_titanic` del esquema `public`, con el fin de predecir la columna `sobrevivio`.

## c45_con_set

### c45_con_set(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, objetos_minimos INTEGER, factor_confianza FLOAT, tabla_pruebas VARCHAR, esquema_pruebas VARCHAR)
Parametros:

- `tabla`: El nombre de la tabla que se usará para crear el arbol
- `atributo`: El nombre del atributo que se va evaluar
- `esquema`: El esquema al cual pertence la tabla que se usará para crear el arbol
- `objetos_minimos`: El minimo número de objetos para no seguir dividiendo una rama en el arbol de desición
- `factor_confianza`: Factor de confianza, se debe especificar un valor entre 0 y 1, un valor mas pequeño llevará a una poda mas agresiva
- `tabla`: El nombre de la tabla a evaluar
- `esquema`: El esquema al cual pertence la tabla que se va evaluar

Ejemplo de uso:

```sql
SELECT * FROM c45("resultado_juego", "jugar", "public", 2, 0.3, "jugar_pruebas", "pruebas")
```

Esta instrucción creará una arbol de desición con el algoritmo c4.5 utilizando la tabla `resultado_juego`, la cual se encuentra en el esquema `public`, con el fin de predecir la columna `jugar`, el minimo de objetos por rama es de `2`, y utiliza un factor de confianza de `0.3`, se evaluará la tabla `jugar_pruebas` en el esquema `pruebas`.

## c45_split

LLama a la función c45, pero divide la tabla en una parte de entrenamiento y otra de pruebas.

### c45_split(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, objetos_minimos INTEGER, factor_confianza FLOAT, porcentaje_entrenamiento INTEGER)

Esta función crea dos tablas nuevas, utiliza una para crear el arbol y la otra para evaluarlo.

Parametros:

- `tabla`: El nombre de la tabla que se usará para crear el arbol y evaluarlo
- `atributo`: El nombre del atributo que se va evaluar
- `esquema`: El esquema al cual pertence la tabla que se usará para crear el arbol y evaluarlo
- `objetos_minimos`: El minimo número de objetos para no seguir dividiendo una rama en el arbol de desición
- `factor_confianza`: Factor de confianza, se debe especificar un valor entre 0 y 1, un valor mas pequeño llevará a una poda mas agresiva
- `porcentaje_entrenamiento`: El porcentaje de la tabla que se va a destinar para entrenamiento (un valor entre 0 y 100)

Ejemplo de uso:

```sql
SELECT * FROM c45_split("sobrevivio", "titanic", "public", 2, 0.3, 70)
```

Esta instrucción creará 2 tablas en el esquema `algoritmos_arbol_datos`, una con el 70% de los datos llamada `public_titanic_70`, y otra con el porcentaje restante llama `public_titanic_30`, la tabla `public_titanic_70` se usará para crear un arbol, con un factor de confiaza de `0.3`, y un minimo de objetos de `2`, luego la tabla `public_titanic_30` se usará para evaluar el arbol

## c45_crossfold_validation

### c45_crossfold_validation(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, objetos_minimos INTEGER, factor_confianza FLOAT, numero_folds INTEGER)

Utiliza crossfold validation para evaluar la tabla

Parametros:

- `tabla`: El nombre de la tabla que se usará para crear el arbol y evaluarlo
- `atributo`: El nombre del atributo que se va evaluar
- `esquema`: El esquema al cual pertence la tabla que se usará para crear el arbol y evaluarlo
- `objetos_minimos`: El minimo número de objetos para no seguir dividiendo una rama en el arbol de desición
- `factor_confianza`: Factor de confianza, se debe especificar un valor entre 0 y 1, un valor mas pequeño llevará a una poda mas agresiva
- `numero_folds`: La cantidad de folds para hacer crossfold validation

Ejemplo de uso:

```sql
SELECT * FROM c45("resultado_juego", "jugar", "public", 2, 0.3, 10)
```

Esta instrucción creará una arbol de desición con el algoritmo c4.5 utilizando la tabla `resultado_juego`, la cual se encuentra en el esquema `public`, con el fin de predecir la columna `jugar`, el minimo de objetos por rama es de `2`, y utiliza un factor de confianza de `0.3`, se evaluará la tabla `jugar` utilizando `10` folds
