from typing import List
import shutil
import os

os.makedirs("./dist/extension", exist_ok=True)

contenido_archivo: str = ""


def crear_funcion_python(
    nombre_funcion: str, nombre_archivo: str, parametros: List[str] = []
):
    print(f"Añadiendo función {nombre_funcion}...")
    global contenido_archivo
    with open(f"./funciones/py/{nombre_archivo}", "r") as script:
        contenido = script.read()
        contenido_archivo += f"""
CREATE OR REPLACE FUNCTION
{nombre_funcion}({", ".join(parametros)})
RETURNS VARCHAR
STRICT
LANGUAGE plpython3u
AS
$$
{contenido}
return 'OK'
$$;
            """


def agregar_archivo(nombre_archivo: str):
    global contenido_archivo

    print(f"Añadiendo contenido de archivo {nombre_archivo}...")
    with open(f"{nombre_archivo}", "r") as archivo:
        contenido = archivo.read()
        contenido_archivo += contenido
        contenido_archivo += "\n"


# Añadir sql
agregar_archivo("./sql/encabezado.sql")
agregar_archivo("./funciones/sql/comunes.sql")
agregar_archivo("./funciones/sql/id3.sql")
agregar_archivo("./funciones/sql/c45.sql")

# Añadir Funciones

crear_funcion_python(
    "dividir_tabla_con_nombre",
    "dividir_tabla_con_nombre.py",
    [
        "tabla VARCHAR",
        "esquema VARCHAR",
        "porcentaje_entrenamiento INTEGER",
        "esquema_entrenamiento VARCHAR",
        "tabla_entrenamiento VARCHAR",
        "esquema_pruebas VARCHAR",
        "tabla_pruebas VARCHAR",
        "sobreescribir BOOLEAN",
    ],
)

crear_funcion_python(
    "dividir_tabla",
    "dividir_tabla.py",
    ["tabla VARCHAR", "esquema VARCHAR", "porcentaje_entrenamiento INTEGER"],
)

crear_funcion_python(
    "crear_arbol_id3",
    "id3.py",
    ["tabla VARCHAR", "atributo_objetivo VARCHAR", "esquema VARCHAR"],
)

crear_funcion_python(
    "crear_arbol_c45",
    "c45.py",
    [
        "tabla VARCHAR",
        "atributo_objetivo VARCHAR",
        "esquema VARCHAR",
        "objetos_minimos INTEGER",
        "factor_confianza FLOAT",
    ],
)

crear_funcion_python(
    "cross_validation",
    "cross_validation.py",
    [
        "tabla VARCHAR",
        "atributo_objetivo VARCHAR",
        "esquema VARCHAR",
        "algoritmo VARCHAR",
        "parametros_adicionales VARCHAR",
        "numero_folds INTEGER",
    ],
)

crear_funcion_python("mostrar_arbol", "mostrar_arbol.py")

crear_funcion_python("mostrar_info_arbol", "mostrar_info_arbol.py")

crear_funcion_python("mostrar_evaluacion", "mostrar_evaluacion.py")

crear_funcion_python("mostrar_info_evaluacion", "mostrar_info_evaluacion.py")

crear_funcion_python(
    "evaluar_matriz",
    "evaluar_matriz.py",
    ["tabla VARCHAR", "esquema VARCHAR", "acumular_valores BOOLEAN"],
)

crear_funcion_python("crear_estadisticas", "crear_estadisticas.py")

crear_funcion_python(
    "predecir_tabla",
    "predecir_tabla.py",
    ["tabla VARCHAR", "esquema VARCHAR", "cantidad_instancias INTEGER"],
)


# Escribir al archivo
print("Escribiendo al archivo algoritmos_arbol--1.0.sql...")
with open("./dist/extension/algoritmos_arbol--1.0.sql", "w") as archivo:
    archivo.write(contenido_archivo)

# Copiar archivo de desinstalación
print("Copiando archivo algoritmos_arbol--uninstall.sql")
shutil.copy("./sql/desinstalar.sql", "./dist/extension/algoritmos_arbol--uninstall.sql")

# Copiar archivo de control
print("Copiando archivo algoritmos_arbol.control")
shutil.copy("./algoritmos_arbol.control", "./dist/extension/algoritmos_arbol.control")

print("Completado")
