The code you've provided is a Python implementation of the C4.5 decision tree algorithm. It doesn't directly use the confidence factor (CF) for pruning in the way the Java implementation does. However, we can add post-pruning to this Python implementation using the confidence factor approach.

Here's how we can modify the code to implement pruning using a confidence factor:

1. First, add a new parameter for the confidence factor:

```python
confidence_factor = 0.25  # Default value, you can adjust this
```

2. Add a new method to the Nodo class for calculating the estimated error:

```python
def estimated_error(self, N, e, CF):
    if CF > 0.5:
        return 0

    if e < 1:
        base = N * (1 - math.pow(CF, 1 / N))
        if e == 0:
            return base
        return base + e * (self.estimated_error(N, 1, CF) - base)

    if e + 0.5 >= N:
        return max(N - e, 0)

    z = Statistics.normal_inverse(1 - CF)
    f = (e + 0.5) / N
    r = (f + (z * z) / (2 * N) + z * math.sqrt((f / N) - (f * f / N) + (z * z / (4 * N * N)))) / (1 + (z * z) / N)

    return (r * N) - e
```

3. Add a method to prune the tree:

```python
def prune(self):
    if self.tipo == "HOJA":
        return

    for hijo in self.hijos:
        hijo.prune()

    if all(hijo.tipo == "HOJA" for hijo in self.hijos):
        error_if_leaf = self.estimated_error(self.cantidad_objetos, self.instancias_cantidad_incorrectas, confidence_factor)
        error_if_not_leaf = sum(hijo.instancias_cantidad_incorrectas for hijo in self.hijos) + \
                            sum(self.estimated_error(hijo.cantidad_objetos, hijo.instancias_cantidad_incorrectas, confidence_factor) for hijo in self.hijos)

        if error_if_leaf <= error_if_not_leaf:
            self.hijos = []
            self.tipo = "HOJA"
            self.dar_clase()
```

4. Modify the main execution to include pruning:

```python
raiz = Nodo(None, "RAIZ", "-1", "=", "-1", atributos)
raiz.crear_hijos()
raiz.prune()  # Add this line to prune the tree after creation
raiz.adicionar_nodos_en_tablas()
```

This implementation adds post-pruning to the tree using the confidence factor approach. It estimates the error for each subtree and compares it with the estimated error if that subtree were replaced by a leaf. If replacing the subtree with a leaf would result in an equal or lower estimated error, the subtree is pruned.

Note that this implementation assumes the existence of a `Statistics.normal_inverse` function, which you may need to implement or import from a statistics library.

Also, keep in mind that this is a basic implementation of pruning and may need further refinement and testing to ensure it works correctly with your specific dataset and requirements.
