# tabla = "test"
# atributo_objetivo = "attr"
# esquema = "public"
# algoritmo = "c45"
# parametros_adicionales = ", 10, 0.25"
# numero_folds = 10

# Limpiar de llamados anteriores:
query = """
    TRUNCATE algoritmos_arbol.matriz_confusion;
    TRUNCATE algoritmos_arbol.clasificados CASCADE;
    TRUNCATE algoritmos_arbol.clasificados_probabilidades;
    """
plpy.execute(query)

plpy.notice("\n====== Crossfold Validation =======\n")

# obtener número de filas por fold
query = f"SELECT count(*) as cantidad FROM {esquema}.{tabla}"
total_filas: int = plpy.execute(query)[0]["cantidad"]
cantidad_iteracion: int = total_filas // numero_folds

restante: int = total_filas - (cantidad_iteracion * numero_folds)

nombre_tabla = f"{esquema}_{tabla}_entrenamiento"

query = f"""
    DROP TABLE IF EXISTS algoritmos_arbol_datos.{nombre_tabla}
"""
plpy.execute(query)

query = f"""
    CREATE TABLE algoritmos_arbol_datos.{nombre_tabla} AS
    SELECT * 
    FROM {esquema}.{tabla}
    ORDER BY random()
"""
plpy.execute(query)

# Añadir id temporal
query = f"""
    ALTER TABLE algoritmos_arbol_datos.{nombre_tabla}
    ADD COLUMN _id_temporal_pruebas INTEGER;
"""
plpy.execute(query)

# Darle valor al id temporal
query = f"""
    WITH numbered_rows AS (
      SELECT *, ROW_NUMBER() OVER () AS row_num, ctid
      FROM algoritmos_arbol_datos.{nombre_tabla}
    )
    UPDATE algoritmos_arbol_datos.{nombre_tabla} AS e
    SET _id_temporal_pruebas = nr.row_num
    FROM numbered_rows nr
    WHERE e.ctid = nr.ctid;
"""
plpy.execute(query)

# ====== Crear matriz de confusion con datos 0 ==========
query = f"""
    SELECT DISTINCT {atributo_objetivo} as clase
    FROM {esquema}.{tabla}
"""

resultado = plpy.execute(query)
clases = [fila["clase"] for fila in resultado]

valores_texto = []
for fila in clases:
    for columna in clases:
        valores_texto.append(f"('{fila}', '{columna}', 0)")

query = f"""
    INSERT INTO algoritmos_arbol.matriz_confusion
    (fila, columna, cantidad)
    VALUES
    {",".join(valores_texto)}
"""
plpy.execute(query)


#  =========== iterar sobre los folds =========
offset = 0

for i in range(numero_folds):
    plpy.notice(f"\n>>> Evaluando fold {i+1}/{numero_folds} <<<")

    query = "DROP TABLE IF EXISTS algoritmos_arbol_datos.fold_prueba"
    plpy.execute(query)
    query = "DROP TABLE IF EXISTS algoritmos_arbol_datos.fold_entrenamiento"
    plpy.execute(query)

    if restante > 0:
        limite = cantidad_iteracion + 1
    else:
        limite = cantidad_iteracion

    query = f"""
        CREATE TABLE algoritmos_arbol_datos.fold_prueba AS
        SELECT * 
        FROM algoritmos_arbol_datos.{nombre_tabla}
        OFFSET {offset}
        LIMIT {limite}
    """
    plpy.execute(query)

    offset += limite
    restante -= 1

    query = f"""
        CREATE TABLE algoritmos_arbol_datos.fold_entrenamiento AS
        SELECT * 
        FROM algoritmos_arbol_datos.{nombre_tabla}
        WHERE _id_temporal_pruebas NOT IN (
            SELECT _id_temporal_pruebas
            FROM algoritmos_arbol_datos.fold_prueba
        )
    """
    plpy.execute(query)

    # Eliminar columna de id temporal
    query = f"""
    ALTER TABLE algoritmos_arbol_datos.fold_prueba
    DROP COLUMN _id_temporal_pruebas
    """
    plpy.execute(query)

    query = f"""
    ALTER TABLE algoritmos_arbol_datos.fold_entrenamiento
    DROP COLUMN _id_temporal_pruebas
    """
    plpy.execute(query)

    # crear arbol
    query = f"""
        SELECT * FROM algoritmos_arbol.crear_arbol_{algoritmo}(
            'fold_entrenamiento',
            '{atributo_objetivo}',
            'algoritmos_arbol_datos'
            {parametros_adicionales}
        )
    """
    plpy.execute(query)

    # añadir a matriz y clasificados
    query = """
        SELECT * 
        FROM algoritmos_arbol.evaluar_matriz(
            'fold_prueba',
            'algoritmos_arbol_datos',
            TRUE
        )
    """
    resultado = plpy.execute(query)


# Insertar como serían las predicciones sin el arbol, luego crear las estadisticas de evaluación

query = f"""
    TRUNCATE TABLE algoritmos_arbol.cantidades_anteriores;
    
    INSERT INTO algoritmos_arbol.cantidades_anteriores
    (
        SELECT
            {atributo_objetivo} as clase,
            count(*) as cantidad
        FROM {esquema}.{tabla}
        GROUP BY {atributo_objetivo}
    );

    SELECT * FROM algoritmos_arbol.crear_estadisticas();
"""
plpy.execute(query)


plpy.notice(f"\n>>> Creando arbol con todos los datos <<<")
# arbol con todos los datos
query = f"""
    SELECT * FROM algoritmos_arbol.crear_arbol_{algoritmo}(
        '{tabla}',
        '{atributo_objetivo}',
        '{esquema}'
        {parametros_adicionales}
    )
"""
plpy.execute(query)


# Limpieza
plpy.execute(f"DROP TABLE IF EXISTS algoritmos_arbol_datos.{nombre_tabla}")
plpy.execute("DROP TABLE IF EXISTS algoritmos_arbol_datos.fold_prueba")
plpy.execute("DROP TABLE IF EXISTS algoritmos_arbol_datos.fold_entrenamiento")
