# tabla = "test"
# esquema = "public"
# acumular_valores = False

if not acumular_valores:
    query = """
        TRUNCATE algoritmos_arbol.matriz_confusion;
        TRUNCATE algoritmos_arbol.clasificados CASCADE;
        TRUNCATE algoritmos_arbol.clasificados_probabilidades;
    """
    plpy.execute(query)

from typing import Dict, List, Literal

class Filtro:
    clase: str
    filtro: str
    probabilidades: Dict[str, float]

    def __init__(self, clase: str, filtro: str, probabilidades: Dict[str, float]):
        self.clase = clase
        self.filtro = filtro
        self.probabilidades = probabilidades

    def __str__(self) -> str:
        return f"""
            <Filtro
                clase: {self.clase}   
                filtro: {self.filtro}   
                probabilidades: {self.probabilidades}   
            >
        """


filtros: List[Filtro] = []

def crear_filtros(id_nodo: int, condicion: str):
    query = f"""
        SELECT tipo, atributo, comparador, atributo_valor
        FROM algoritmos_arbol.nodos
        WHERE id = {id_nodo};
    """
    resultado = plpy.execute(query)[0]

    atributo = resultado["atributo"]
    comparador = resultado["comparador"]
    atributo_valor = resultado["atributo_valor"]

    condicion = f"{condicion} AND {atributo} {comparador} '{atributo_valor}'"

    tipo: Literal["RAIZ", "RAMA", "HOJA"] = resultado["tipo"]

    if tipo == "HOJA":
        query = f"""
            SELECT 
                clase,
                instancias_cantidad_correctas,
                instancias_cantidad_incorrectas
            FROM algoritmos_arbol.nodos
            WHERE id = {id_nodo};
        """
        clase = plpy.execute(query)[0]["clase"]

        query = f"""
            SELECT clase, probabilidad
            FROM algoritmos_arbol.clases_probabilidades
            WHERE nodo = {id_nodo}
        """

        resultado = plpy.execute(query)

        probabilidades: Dict[str, float] = {
            fila["clase"]: fila["probabilidad"] for fila in resultado
        }

        filtro = Filtro(clase, condicion, probabilidades)

        filtros.append(filtro)

        return

    query = f"""
        SELECT id
        FROM algoritmos_arbol.nodos
        WHERE id_padre = {id_nodo};
    """
    resultado = plpy.execute(query)

    for fila in resultado:
        crear_filtros(fila["id"], condicion)


query = """
    SELECT valor as atributo_objetivo
    FROM algoritmos_arbol.info
    WHERE id = 'attr_objetivo';
    """
resultado = plpy.execute(query)[0]
atributo_objetivo: str = resultado["atributo_objetivo"]


query = f"""
    SELECT DISTINCT {atributo_objetivo}::text AS clase
    FROM {esquema}.{tabla}
    UNION
    SELECT DISTINCT clase::text
    FROM algoritmos_arbol.nodos
    WHERE clase IS NOT NULL
    """
resultado = plpy.execute(query)
matriz: Dict[str, Dict[str, int]] = {}

# Crear matriz
for fila in resultado:
    clases_matriz = {}
    for fila_interno in resultado:
        clases_matriz[fila_interno["clase"]] = 0

    matriz[fila["clase"]] = clases_matriz


query = f"""
    SELECT id
    FROM algoritmos_arbol.nodos
    WHERE id_padre = 0;
"""
resultado = plpy.execute(query)

if len(resultado) == 0:
    query = f"""
        SELECT clase
        FROM algoritmos_arbol.nodos
        WHERE tipo = 'RAIZ'
    """
    resultado = plpy.execute(query)[0]

    clase = resultado["clase"]

    query = f"""
        SELECT clase, probabilidad
        FROM algoritmos_arbol.clases_probabilidades
        WHERE nodo = 0
    """

    resultado = plpy.execute(query)

    probabilidades: Dict[str, float] = {
        fila["clase"]: fila["probabilidad"] for fila in resultado
    }

    filtro = Filtro(clase, "TRUE", probabilidades)

    filtros.append(filtro)
else:
    for fila in resultado:
        crear_filtros(fila["id"], "true")



for filtro in filtros:
    query = f"""
        SELECT
            count(*) as cuenta,
            {atributo_objetivo}::text as clase
        FROM {esquema}.{tabla}
        WHERE {filtro.filtro}
        GROUP BY {atributo_objetivo}
    """
    resultado = plpy.execute(query)

    for fila in resultado:
        query = f"""
            INSERT INTO algoritmos_arbol.clasificados
            (clase, cantidad)
            VALUES ('{fila["clase"]}', {fila["cuenta"]})
            RETURNING id;
        """
        id = plpy.execute(query)[0]["id"]

        probabilidades_texto: List[str] = []
        for clase, valor in filtro.probabilidades.items():
            probabilidades_texto.append(f"({id}, '{clase}', {valor})")


        query = f"""
            INSERT INTO algoritmos_arbol.clasificados_probabilidades
            (id, clase, probabilidad)
            VALUES
            {",".join(probabilidades_texto)}
        """

        plpy.execute(query)

        clases_matriz = matriz[fila["clase"]]
        clases_matriz[filtro.clase] += fila["cuenta"]


for fila, clase_clasificada in matriz.items():
    for columna, cantidad in clase_clasificada.items():
        if acumular_valores:
            query = f"""
                UPDATE algoritmos_arbol.matriz_confusion
                    SET cantidad = cantidad + {cantidad}
                WHERE 
                    columna = '{columna}'
                    AND fila = '{fila}'
            """
        else:
            query = """
                INSERT INTO algoritmos_arbol.matriz_confusion
                ( fila, columna, cantidad)
                VALUES ('%s', '%s', %s)
            """ % (
                fila,
                columna,
                cantidad,
            )

        plpy.execute(query)

# Insertar como serían las predicciones sin el arbol
query = f"""
    TRUNCATE TABLE algoritmos_arbol.cantidades_anteriores;
    INSERT INTO algoritmos_arbol.cantidades_anteriores
    (
        SELECT
            {atributo_objetivo} as clase,
            count(*) as cantidad
        FROM {esquema}.{tabla}
        GROUP BY {atributo_objetivo}
    )
"""
plpy.execute(query)
