from collections import defaultdict
from typing import Dict

# Crear tabla
headers = {
    "Clase": "clase",
    "TPR": "tpr",
    "FPR": "fpr",
    "Precision": "presicion",
    "Recall": "recall",
    "F-Measure": "f_measure",
    "MCC": "mcc",
    "ROC-Area": "roc_area",
    "PRC-Area": "prc_area",
}
table_data = []

query = f"""
    SELECT tipo, {", ".join(headers.values())}
    FROM algoritmos_arbol.estadisticas_evaluacion
"""

resultado = plpy.execute(query)

for fila in resultado:
    row = []

    for columna in headers.values():
        value = fila[columna]

        if value is None:
            row.append("-")
        elif isinstance(value, float):
            row.append(f"{value:.3f}")
        else:
            row.append(str(value))
    table_data.append(row)

max_widths = []
for i in range(len(headers)):
    max_length = 0
    for row in table_data + [list(headers.values())]:
        length = len(str(row[i]))
        if length > max_length:
            max_length = length
    max_widths.append(max_length)

table_str = "==== Presición detallada por clase ====\n"

table_str += (
    "  ".join(f"{header:<{max_widths[i]}}" for i, header in enumerate(headers)) + "\n"
)

for row in table_data:
    table_str += (
        "  ".join(f"{str(item):<{max_widths[i]}}" for i, item in enumerate(row)) + "\n"
    )

table_str += "\n"

plpy.execute("delete from algoritmos_arbol.textos_evaluacion where id='presicion_clases'")
plpy.execute("insert into algoritmos_arbol.textos_evaluacion(id, valor) values ('presicion_clases', '{}')".format(table_str))

plpy.notice(table_str)


# string para mostrar la matriz
def nombre_columna(indice: int) -> str:
    indice += 1
    resultado = ""
    while indice > 0:
        indice -= 1
        resultado = chr(97 + (indice % 26)) + resultado
        indice //= 26
    return resultado


matriz: Dict[str, Dict[str, int]] = defaultdict(dict)

# Crear matriz
query = """
    SELECT fila, columna, cantidad
    FROM algoritmos_arbol.matriz_confusion
    ORDER BY fila, columna
"""

resultado = plpy.execute(query)

for fila_resultado in resultado:
    fila: str = fila_resultado["fila"]
    columna: str = fila_resultado["columna"]
    cantidad: int = fila_resultado["cantidad"]

    matriz[fila][columna] = cantidad

cantidad_maxima = max(max(matriz_filas.values()) for matriz_filas in matriz.values())
longitud = len(str(cantidad_maxima)) + 1

matriz_str = "==== Matriz de Confusión ====\n"
for indice in range(len(matriz)):
    matriz_str += nombre_columna(indice).ljust(longitud)

matriz_str += "  <-- Clasificado como"

for indice, (clase, clase_clasificada) in enumerate(matriz.items()):
    matriz_str += "\n"
    for cantidad in clase_clasificada.values():
        cantidad_str = str(cantidad)
        matriz_str += cantidad_str.ljust(longitud)
    matriz_str += "| "
    matriz_str += nombre_columna(indice)
    matriz_str += " = "
    matriz_str += clase

matriz_str += "\n"

plpy.execute("delete from algoritmos_arbol.textos_evaluacion where id='matriz'")
plpy.execute("insert into algoritmos_arbol.textos_evaluacion(id, valor) values ('matriz', '{}')".format(matriz_str.replace("'", "''")))

plpy.notice(matriz_str)
