# esquema = "public"
# tabla = "test"
# cantidad_instancias = 100
from typing import List, Literal

query = """
    SELECT valor as atributo_objetivo
    FROM algoritmos_arbol.info
    WHERE id = 'attr_objetivo';
    """
resultado = plpy.execute(query)[0]
atributo_objetivo: str = resultado["atributo_objetivo"]


class Condicion:
    atributo: str
    comparador: str
    atributo_valor: str

    def __init__(self, atributo: str, comparador: str, atributo_valor: str) -> None:
        self.atributo = atributo
        self.comparador = comparador
        self.atributo_valor = atributo_valor


class Filtro:
    condiciones: List[Condicion]
    clase: str
    certeza: float

    def __init__(self) -> None:
        self.condiciones = []
        self.clase = ""
        self.certeza = 0

    def obtener_condicion_sql(self) -> str:
        sql = "true"
        for condicion in self.condiciones:
            sql += f" and {condicion.atributo} {condicion.comparador} '{condicion.atributo_valor}'"

        return sql

    def copiar_filtro(self) -> "Filtro":
        nuevo = Filtro()
        nuevo.condiciones = self.condiciones[::]
        return nuevo


filtros: List[Filtro] = []


def crear_filtros(id_nodo: int, filtro: Filtro):
    query = f"""
        SELECT tipo, atributo, comparador, atributo_valor
        FROM algoritmos_arbol.nodos
        WHERE id = {id_nodo};
    """
    resultado = plpy.execute(query)[0]

    atributo = resultado["atributo"]
    comparador = resultado["comparador"]
    atributo_valor = resultado["atributo_valor"]

    condicion = Condicion(atributo, comparador, atributo_valor)

    filtro.condiciones.append(condicion)

    tipo: Literal["RAIZ", "RAMA", "HOJA"] = resultado["tipo"]

    if tipo == "HOJA":
        query = f"""
            SELECT clase,
            instancias_cantidad_correctas,
            instancias_cantidad_incorrectas
            FROM algoritmos_arbol.nodos
            WHERE id = {id_nodo};
        """
        resultado = plpy.execute(query)[0]
        ins_correctas = resultado["instancias_cantidad_correctas"]
        ins_incorrectas = resultado["instancias_cantidad_incorrectas"]

        filtro.clase = resultado["clase"]
        total = ins_correctas + ins_incorrectas

        if total == 0:
            filtro.certeza = 1
        else:
            filtro.certeza = ins_correctas / total

        filtros.append(filtro)

    query = f"""
        SELECT id
        FROM algoritmos_arbol.nodos
        WHERE id_padre = {id_nodo};
    """
    resultado = plpy.execute(query)

    for fila in resultado:
        crear_filtros(fila["id"], filtro.copiar_filtro())


query = f"""
    SELECT id
    FROM algoritmos_arbol.nodos
    WHERE id_padre = 0;
"""
resultado = plpy.execute(query)

if len(resultado) == 0:
    query = f"""
        SELECT clase
        FROM algoritmos_arbol.nodos
        WHERE tipo = 'RAIZ'
        instancias_cantidad_correctas,
        instancias_cantidad_incorrectas
        FROM algoritmos_arbol.nodos;
        """
    resultado = plpy.execute(query)[0]
    ins_correctas = resultado["instancias_cantidad_correctas"]
    ins_incorrectas = resultado["instancias_cantidad_incorrectas"]

    filtro = Filtro()
    filtro.clase = resultado["clase"]
    filtro.certeza = ins_correctas / (ins_correctas + ins_incorrectas)

    resultado = plpy.execute(query)
    filtros.append(filtro)
else:
    for fila in resultado:
        crear_filtros(fila["id"], Filtro())

# Tabla donde se guardan las predicciones
query = f"""
DROP TABLE IF EXISTS algoritmos_arbol_datos.predicciones;
"""
plpy.execute(query)

query = f"""
CREATE TABLE algoritmos_arbol_datos.predicciones
AS TABLE {esquema}.{tabla};
"""
plpy.execute(query)

query = f"""
ALTER TABLE algoritmos_arbol_datos.predicciones
ADD COLUMN _actual TEXT DEFAULT NULL,
ADD COLUMN _predicho TEXT DEFAULT NULL,
ADD COLUMN _error BOOL DEFAULT FALSE,
ADD COLUMN _prediccion FLOAT DEFAULT NULL;
"""
plpy.execute(query)

for filtro in filtros:
    query = f"""
    UPDATE algoritmos_arbol_datos.predicciones
    SET _actual = {atributo_objetivo},
        _predicho = '{filtro.clase}',
        _prediccion = {filtro.certeza}
    WHERE {filtro.obtener_condicion_sql()};
    """
    plpy.execute(query)

query = f"""
UPDATE algoritmos_arbol_datos.predicciones
SET _error = true
WHERE not _actual = _predicho;
"""
plpy.execute(query)

# Resumen

query = f"""
    SELECT count(*) AS cantidad
    FROM algoritmos_arbol_datos.predicciones
    WHERE _predicho is not NULL
"""
predichos = plpy.execute(query)[0]["cantidad"]

query = f"""
    SELECT count(*) AS cantidad
    FROM algoritmos_arbol_datos.predicciones
    WHERE _predicho is NULL
"""
no_predichos = plpy.execute(query)[0]["cantidad"]

query = f"""
    SELECT count(*) AS cantidad
    FROM algoritmos_arbol_datos.predicciones
    WHERE _actual = _predicho
    AND _actual is not NULL
    AND _predicho is not NULL
"""
predichos_correctamente = plpy.execute(query)[0]["cantidad"]

query = f"""
    SELECT count(*) AS cantidad
    FROM algoritmos_arbol_datos.predicciones
    WHERE _actual != _predicho
    AND _actual is not NULL
    AND _predicho is not NULL
"""
predichos_incorrectamente = plpy.execute(query)[0]["cantidad"]

texto = "==== Resumen de Predicciones ====\n"
texto += f"Instancias predichas: {predichos}\n"
texto += f"Instancias no predichas: {no_predichos}\n"
texto += f"Instancias predichas correctamente: {predichos_correctamente}\n"
texto += f"Instancias predichas incorrectamente: {predichos_incorrectamente}\n"
texto += "\n"

plpy.execute(
    "delete from algoritmos_arbol.textos_evaluacion where id='resumen_prediccion'"
)
plpy.execute(
    "insert into algoritmos_arbol.textos_evaluacion(id, valor) values ('arbol', '{}')".format(
        texto.replace("'", "''")
    )
)
plpy.notice(texto)


if cantidad_instancias != 0:
    if cantidad_instancias < 0:
        limite = ""
    else:
        limite = f"LIMIT {cantidad_instancias}"
    # Mostrar predicciones
    query = f"""
        SELECT DISTINCT {atributo_objetivo} AS clase
        FROM algoritmos_arbol_datos.predicciones
        UNION
        SELECT DISTINCT clase
        FROM algoritmos_arbol.nodos
        WHERE clase IS NOT NULL
        """
    resultado = plpy.execute(query)
    clases = [fila["clase"] for fila in resultado if fila["clase"] is not None]
    clase_mas_larga = max(clases, key=len)

    query = "SELECT count(*) as total FROM algoritmos_arbol_datos.predicciones"
    resultado = plpy.execute(query)

    pad_instancia = len(str(resultado[0]["total"]))
    pad_instancia = max(len("Instancia#"), pad_instancia)
    pad_actual = max(len("Actual"), len(clase_mas_larga))
    pad_predicho = max(len("Predicho"), len(clase_mas_larga))

    texto = "==== Predicciones ====\n"

    texto += f"{"Instancia#".ljust(pad_instancia)}  {"Actual".ljust(pad_actual)}  {"Predicho".ljust(pad_predicho)}  Error  Predicción\n"

    query = f"""
        SELECT _actual, _predicho, _error, _prediccion
        FROM algoritmos_arbol_datos.predicciones
        {limite}
    """

    cursor = plpy.cursor(query)

    inst = 1
    while True:
        filas = cursor.fetch(5000)
        if not filas:
            break

        resultado = ""
        for fila in filas:
            actual: str = fila["_actual"]
            if actual is None:
                actual = "?"

            predicho: str = fila["_predicho"]
            if predicho is None:
                predicho = "?"

            error: bool = fila["_error"]

            prediccion: float | str = fila["_prediccion"]
            if prediccion is None:
                prediccion = "?"

            resultado += f"{str(inst).ljust(pad_instancia)}  {actual.ljust(pad_actual)}  {predicho.ljust(pad_predicho)}  {str(error).ljust(5)}  {prediccion:.3f}\n"
            inst += 1

        texto += resultado

    texto += "\n"
    plpy.execute(
        "delete from algoritmos_arbol.textos_evaluacion where id='instancias_predichas'"
    )
    plpy.execute(
        "insert into algoritmos_arbol.textos_evaluacion(id, valor) values ('arbol', '{}')".format(
            texto.replace("'", "''")
        )
    )
    plpy.notice(texto)
