CREATE OR REPLACE FUNCTION c45(tabla VARCHAR, atributo VARCHAR)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
BEGIN
    PERFORM algoritmos_arbol.c45(tabla, atributo, 'public', 1, 0.25::float);

    RETURN 'OK'::VARCHAR;
END
$$;

CREATE OR REPLACE FUNCTION c45(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, objetos_minimos INTEGER, factor_confianza FLOAT)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
BEGIN
  PERFORM algoritmos_arbol.crear_arbol_c45(tabla, atributo, esquema, objetos_minimos, factor_confianza);

  PERFORM algoritmos_arbol.evaluar_tabla(tabla, esquema);

  PERFORM algoritmos_arbol.mostrar();

  RETURN 'OK'::VARCHAR;
END
$$;


CREATE OR REPLACE FUNCTION c45_con_set(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, objetos_minimos INTEGER, factor_confianza FLOAT, tabla_pruebas VARCHAR, esquema_pruebas VARCHAR)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
BEGIN
    PERFORM algoritmos_arbol.crear_arbol_c45(tabla, atributo, esquema, objetos_minimos, factor_confianza);

    PERFORM algoritmos_arbol.evaluar_tabla(tabla_pruebas, esquema_pruebas);

    PERFORM algoritmos_arbol.mostrar();

    RETURN 'OK'::VARCHAR;
END
$$;


CREATE OR REPLACE FUNCTION c45_crossfold_validation(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, objetos_minimos INTEGER, factor_confianza FLOAT, numero_folds INTEGER)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
DECLARE
    parametros TEXT;
BEGIN
    parametros := format(', %s, %s', objetos_minimos, factor_confianza);

    PERFORM algoritmos_arbol.cross_validation(tabla, atributo, esquema, 'c45', parametros, numero_folds);

    PERFORM algoritmos_arbol.mostrar();

    RETURN 'OK'::VARCHAR;
END
$$;


CREATE OR REPLACE FUNCTION c45_split(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, objetos_minimos INTEGER, factor_confianza FLOAT, porcentaje_entrenamiento INTEGER)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
DECLARE
    entrenamiento TEXT;
    pruebas TEXT;
BEGIN
    entrenamiento := format('%s_%s_e%s', esquema, tabla, porcentaje_entrenamiento);
    pruebas := format('%s_%s_p%s', esquema, tabla, 100 - porcentaje_entrenamiento);

    PERFORM algoritmos_arbol.dividir_tabla(tabla, esquema, porcentaje_entrenamiento);

    PERFORM algoritmos_arbol.crear_arbol_c45(entrenamiento, atributo, 'algoritmos_arbol_datos', objetos_minimos, factor_confianza);

    PERFORM algoritmos_arbol.evaluar_tabla(pruebas, 'algoritmos_arbol_datos');

    PERFORM algoritmos_arbol.mostrar();

    RETURN 'OK'::VARCHAR;
END
$$;
