CREATE OR REPLACE FUNCTION id3(tabla VARCHAR, atributo VARCHAR)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
BEGIN
    PERFORM algoritmos_arbol.id3(tabla, atributo, 'public');

    RETURN 'OK'::VARCHAR;
END
$$;

CREATE OR REPLACE FUNCTION id3(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
BEGIN
    PERFORM algoritmos_arbol.crear_arbol_id3(tabla, atributo, esquema);

    PERFORM algoritmos_arbol.evaluar_tabla(tabla, esquema);

    PERFORM algoritmos_arbol.mostrar();

    RETURN 'OK'::VARCHAR;
END
$$;


CREATE OR REPLACE FUNCTION id3_con_set(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, tabla_pruebas VARCHAR, esquema_pruebas VARCHAR)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
BEGIN
    PERFORM algoritmos_arbol.crear_arbol_id3(tabla, atributo, esquema);

    PERFORM algoritmos_arbol.evaluar_tabla(tabla_pruebas, esquema_pruebas);

    PERFORM algoritmos_arbol.mostrar();

    RETURN 'OK'::VARCHAR;
END
$$;


CREATE OR REPLACE FUNCTION id3_crossfold_validation(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR, numero_folds INTEGER)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
DECLARE
    parametros TEXT;
BEGIN
    parametros := ''

    PERFORM algoritmos_arbol.cross_validation(tabla, atributo, esquema, 'id3', parametros, numero_folds);

    PERFORM algoritmos_arbol.mostrar();

    RETURN 'OK'::VARCHAR;
END
$$;


CREATE OR REPLACE FUNCTION id3_split(tabla VARCHAR, atributo VARCHAR, esquema VARCHAR,  porcentaje_entrenamiento INTEGER)
    RETURNS VARCHAR
    STRICT
    LANGUAGE plpgsql
AS
$$
DECLARE
    entrenamiento TEXT;
    pruebas TEXT;
BEGIN
    entrenamiento := format('%s_%s_e%s', esquema, tabla, porcentaje_entrenamiento);
    pruebas := format('%s_%s_p%s', esquema, tabla, 100 - porcentaje_entrenamiento);

    PERFORM algoritmos_arbol.dividir_tabla(tabla, esquema, porcentaje_entrenamiento);

    PERFORM algoritmos_arbol.crear_arbol_id3(entrenamiento, atributo, 'algoritmos_arbol_datos');

    PERFORM algoritmos_arbol.evaluar_tabla(pruebas, 'algoritmos_arbol_datos');

    PERFORM algoritmos_arbol.mostrar();

    RETURN 'OK'::VARCHAR;
END
$$;
