use tokio::sync::Mutex;

use tauri::Manager;

mod database;

#[cfg_attr(mobile, tauri::mobile_entry_point)]
pub fn run() {
    tauri::Builder::default()
        .plugin(tauri_plugin_dialog::init())
        .plugin(tauri_plugin_fs::init())
        .setup(|app| {
            if cfg!(debug_assertions) {
                app.handle().plugin(
                    tauri_plugin_log::Builder::default()
                        .level(log::LevelFilter::Info)
                        .build(),
                )?;
            };
            app.manage(Mutex::new(database::ConnectionManager::default()));
            Ok(())
        })
        .invoke_handler(tauri::generate_handler![
            database::database_connect,
            database::database_disconnect,
            database::test_connection,
            database::get_table_row_count,
            database::divide_table_manual,
            database::get_schemas,
            database::get_tables,
            database::get_columns,
            database::cancel_query,
            database::get_column_data_distribution,
            database::cut_table_columns,
            database::tree::evaluation_training_table,
            database::tree::evaluation_specific_table,
            database::tree::predict_table_values,
            database::tree::evaluation_crossfold_validation,
            database::tree::evaluation_percentage_split,
            database::csv::get_csv_first_rows,
            database::csv::table_from_file,
            database::clustering::load_table_for_kmeanspy,
            database::clustering::kmeans_main,
            database::clustering::kmeans_elbow,
            database::clustering::kmeans_silhouette,
            database::clustering::kmeans_summary,
            database::clustering::kmeans_centroids,
            database::clustering::kmeans_results
        ])
        .run(tauri::generate_context!())
        .expect("error while running tauri application");
}
