import { readJsonFromFile, writeJsonToFile } from './io';
import type { ConnectionDetails } from './types';
import { invoke } from '@tauri-apps/api/core';

export const getConnections = async (): Promise<ConnectionDetails[]> => {
	try {
		const connections = await readJsonFromFile<ConnectionDetails[]>('connections.json');
		return connections;
	} catch (error) {
		throw error;
	}
};

export const saveConnections = async (connections: ConnectionDetails[]) => {
	await writeJsonToFile('connections.json', connections);
};

export const getSchemas = async (): Promise<string[]> => {
	const result = await invoke('get_schemas');
	return result as string[];
};

export const getTables = async (schema: string): Promise<string[]> => {
	const result = await invoke('get_tables', { schema });
	return result as string[];
};

export const getColumns = async (schema: string, table: string): Promise<string[]> => {
	const result = await invoke('get_columns', { schema, table });
	return result as string[];
};

export const cancelQuery = async (connection: ConnectionDetails | null, pid: number) => {
	if (!connection) return;

	await invoke('cancel_query', { connectionDetails: connection, pid: pid });
};
