import { BaseDirectory, mkdir, readTextFile, writeTextFile } from '@tauri-apps/plugin-fs';

export const writeJsonToFile = async <T>(filename: string, data: T): Promise<void> => {
	try {
		const jsonString = JSON.stringify(data, null, 2);
		await writeTextFile(`${filename}`, jsonString, { baseDir: BaseDirectory.AppData });
	} catch (error) {
		console.error('Error writing file:', error);
		throw error;
	}
};

export const readJsonFromFile = async <T>(filename: string): Promise<T> => {
	try {
		const jsonString = await readTextFile(`${filename}`, {
			baseDir: BaseDirectory.AppData
		});
		return JSON.parse(jsonString) as T;
	} catch (error) {
		console.error('Error reading file:', error);
		throw error;
	}
};
