import { writable } from 'svelte/store';
import type { ConnectionDetails } from './types';
import { invoke } from '@tauri-apps/api/core';

export const connectionDetails = writable<ConnectionDetails | null>(null);
export const busyRefCount = writable<number>(0);
export const loadingMessage = writable<string>('');

export const connectionPID = writable<number>(0);

export const lockApp = () => {
	busyRefCount.update((count) => count + 1);
};

export const unlockApp = () => {
	busyRefCount.update((count) => count - 1);
};

connectionDetails.subscribe(async (conn) => {
	if (conn === null) {
		lockApp();
		try {
			await invoke('database_disconnect');
			console.log('Desconectando base de datos...');
		} catch (e) {
			console.log(e);
		}
		unlockApp();
	}
});
